# -*- coding: utf-8 -*-
"""
Created on Mon Mar  4 20:51:22 2019

@author: yoelr
"""
from .. import Unit
from .decorators import cost
from .metaclasses import splitter

@cost('Area', units='ft^2', ub=200, CE=567, cost=1010, n=0.91, BM=1.73,
      fsize=lambda self: self.ins[0].massnet/(self.capacity*self.mesh_opening))
class VibratingScreen(Unit, metaclass=splitter):
    # Assume 3-deck vibrating screen
    
    #: Flow rate per area of screen per apareture (kg∕ft2-hr-mm)
    capacity = 0.2*907.18474
    
    #: Mesh opening (mm)
    mesh_opening = 8


# -*- coding: utf-8 -*-
"""
Created on Wed May  1 19:05:53 2019

@author: yoelr
"""
from .._hx import HXutility as _HXutility

__all__ = ('attach_heat_exchanger',)

def attach_heat_exchanger(cls):
    if cls._N_heat_utilities != 0:
        raise ValueError("number of heat utilities (_N_heat_utilities) must be 0 to attach heat exchanger")
    
    class NewUnit(cls):
        line = cls.line
            
        def _init(self):
            super()._init()
            self._heat_exchanger = he = _HXutility(None, None) 
            self._heat_utilities = he._heat_utilities
            he._ins = self._ins
            he._outs = self._outs
        
        def _design(self):
            super()._design()
            self._heat_exchanger._design()
            
        def _cost(self):
            super()._cost()
            he = self._heat_exchanger
            he._cost()
            self._Cost['Heat exchanger'] = he._Cost['Heat exchanger'] 
        
        _init.__doc__ = cls.__doc__
        _design.__doc__ = cls.__doc__
        _cost.__doc__ = cls.__doc__
    
    NewUnit.__name__ = cls.__name__
    return NewUnit
 


# -*- coding: utf-8 -*-
"""
Created on Tue May 28 11:32:33 2019

@author: yoelr
"""

__all__ = []

from ._boiler_turbogenerator import *
from ._cooling_tower import *

from . import _boiler_turbogenerator
from . import _cooling_tower

__all__.extend(_boiler_turbogenerator.__all__)
__all__.extend(_cooling_tower.__all__)


# -*- coding: utf-8 -*-
"""
Created on Wed Dec  5 16:42:49 2018

This module includes classes and functions relating arrays.

@author: Guest Group
"""
import numpy as np
from array_collections import PropertyFactory, tuple_array, property_array

__all__ = ('tuple_array', 'property_array',
           'PropertyFactory', 'fraction')

ndarray = np.ndarray
asarray = np.asarray

# %% Functions

# def reorder(array: 'array_like', current_order: tuple, new_order: tuple) -> np.ndarray:
#     """Return a reordered array with zero in place of elements that exist in the new order but not in the current order.
    
#     .. code-block:: python
       
#        >>> reorder([1,2,3], ('a', 'b', 'c'), ('a', 'c', 'd', 'b'))
#        array([1, 3, 0, 2])
    
#     """
#     n = 0
#     p_new = np.zeros(len(new_order), dtype=float)
#     for ID1 in current_order:
#         m = 0
#         for ID2 in new_order:
#             if ID1 is ID2:
#                 p_new[m] = array[n]
#             m += 1
#         n += 1
#     return p_new

def fraction(array):
    """Return a normalized array to a magnitude of 1. If magnitude is zero, all fractions will have equal value."""
    sum_array = array.sum()
    if sum_array == 0:
        l = len(array)
        return np.ones(l)/l
    else:
        return array/sum_array
    
# def same_type(iterable, type_):
#     """Raise TypeError if any item in iterable is not an instance of type_."""
#     # Make sure iterable is in a tuple
#     if not isinstance(iterable, type_):
#         # Check that all are type_ instances
#         for s in iterable:
#             if not isinstance(s, type_):
#                 raise TypeError(f"Only '{type_.__name__}' objects are valid elements, not '{type(s).__name__}' objects.")


# -*- coding: utf-8 -*-
"""
Created on Mon Mar  4 11:10:49 2019

@author: yoelr
"""
from .. import Unit
from .decorators import cost
from .metaclasses import static

@cost('Flow rate', CE=567, cost=813, ub=500, n=0.38)
class ConveyingBelt(Unit, metaclass=static):
    length = 40 #: ft
    height = 20 #: ft
    _N_outs = 1
    _has_power_utility = True
    _minimum_flow = 120
    _units = {'Flow rate': 'ft^3/hr'}
    
    def _design(self):
        feed = self.ins[0]
        self._Design['Flow rate'] = volnet = feed.volnet*35.315 # ft3/hr
        if volnet < self._minimum_flow:
            self._lb_warning('Flow rate', volnet, self._minimum_flow)
        massnet = feed.massnet*0.0006124 #lb/s
        self._power_utility(0.00058*massnet**0.82*self.length + self.height*0.00182*massnet * 0.7457) # kW
        
        




# -*- coding: utf-8 -*-
"""
Created on Sat Aug 18 14:36:58 2018

@author: yoelr
"""
import copy

__all__ = ('Graphics',)


# %%

_node_function = lambda unit: None
_edge_in = [{'headport': 'c'} for i in range(3)]
_edge_out = [{'tailport': 'c'} for i in range(2)]
_node = {'shape': 'box',
         'fillcolor': "white:#CDCDCD",
         'style': 'filled',
         'gradientangle': '0',
         'width': '0.6',
         'height': '0.6',
         'orientation': '0.0',
         'color': 'black',
         'peripheries': '1',
         'margin': 'default'}

# %%
    
class Graphics:
    """Create a Graphics object that contains specifications for Graphviz node and edge styles"""
    def __init__(self, edge_in=None,
                 edge_out=None,
                 node=None,
                 node_function=None,
                 in_system=True):
        # [dict] Input stream edge settings
        self.edge_in = edge_in or copy.deepcopy(_edge_in)
        
        # [dict] Output stream edge settings
        self.edge_out = edge_out or copy.deepcopy(_edge_out)
        
        #: [dict] Node settings
        self.node = node or copy.copy(_node)
        
        #: [function] Is called to update node settings
        self.node_function = node_function or _node_function
        
        #: [bool] True for Unit object to appear within a system diagram
        self.in_system = in_system

    def __repr__(self):
        return 'f<{type(self).__name__}>'

default_graphics = Graphics()



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 21:23:56 2018

@author: yoelr
"""
from .. import Unit, MixedStream
from ._flash import RatioFlash
from .metaclasses import splitter
from .decorators import cost

__all__ = ('Centrifuge_LLE', 'RatioCentrifuge_LLE', 'SplitCentrifuge_LLE')

# electricity kW/(m3/hr) from USDA biosdiesel Super Pro model
# Possibly 1.4  kW/(m3/hr)
# https://www.sciencedirect.com/topics/engineering/disc-stack-centrifuge
# Microalgal fatty acids—From harvesting until extraction H.M. Amaro, et. al.,
# in Microalgae-Based Biofuels and Bioproducts, 2017

@cost('Flow rate', units='m^3/hr', CE=525.4, cost=28100,
      n=0.574, kW=3.66, ub=100, BM=2.03)
class Centrifuge_LLE(Unit):
    r"""Create an equlibrium based centrifuge with the option of having liquid non-keys and LIQUID non-keys completly separate into their respective phases.

    Equation for f.o.b cost:

    :math:`C_{fob}^{2007} = 28100 Q^{0.574} (0.1 < Q < 100 \frac{m^3}{h})` 

    **Parameters**

        **species_IDs:** *tuple[str]* IDs of equilibrium species
    
        **split:** *tuple[float]* Initial guess split fractions of equilibrium specise to the 'liquid' phase.
    
        **lNK:** *tuple[str]* Species assumed to completely remain in the 'liquid' phase.
    
        **LNK:** *tuple[str]* Species assumed to completely remain in the 'LIQUID' phase.
    
        **solvents:** *tuple[str]* Species corresponding to specified solvent_split.
    
        **solvent_split:** *tuple[float]* Split fractions of each specie to the 'liquid' phase.
         
    **ins**
    
        [0] Input stream
        
    **outs**
    
        [0] 'liquid' phase stream
        
        [1] 'LIQUID' phase stream
    
    **Examples**
    
        :doc:`Centrifuge_LLE Example`
    
    """
    line = 'Liquids Centrifuge'
    _kwargs = {'species_IDs': None,
               'split': None,
               'lNK': (),
               'LNK': (),
               'solvents': (),
               'solvent_split': ()}
    _bounds = {'Flow rate': (0.1, 100)}

    def _init(self):
        self._mixedstream =  MixedStream(None)

    def _run(self):
        liq, LIQ = self.outs
        feed = self.ins[0]
        ms = self._mixedstream
        ms.empty()
        ms.liquid_mol[:] = feed.mol
        ms.LLE(T=feed.T,  **self._kwargs)
        liq._mol[:] = ms.liquid_mol
        LIQ._mol[:] = ms.LIQUID_mol
        liq.T = LIQ.T = ms.T
        liq.P = LIQ.P = ms.P
        
# class PartitionCentrifuge_LLE(Centrifuge_LLE):
#     _N_heat_utilities = 0
#     kwargs = PartitionFlash._kwargs
#     _run = PartitionFlash._run 
#     _setup = PartitionFlash._setup
#     def _set_phases(self):
#         top, bot = self.outs
#         top.phase = 'l'
#         bot.phase = 'L'

class RatioCentrifuge_LLE(Centrifuge_LLE):
    _N_heat_utilities = 0
    _kwargs = {'Kspecies': [],  # list of species that correspond to Ks
               'Ks': [],  # list of molar ratio partition coefficinets,
               # Ks = y/x, where y and x are molar ratios of two different phases
               'top_solvents': [],  # list of species that correspond to top_split
               'top_split': [],  # list of splits for top_solvents
               'bot_solvents': [],  # list of species that correspond to bot_split
               'bot_split': []}  # list of splits for bot_solvents
    _run = RatioFlash._run
    
class SplitCentrifuge_LLE(Centrifuge_LLE, metaclass=splitter): pass
    
    
    


# -*- coding: utf-8 -*-
"""
Created on Tue May 14 14:35:55 2019

@author: yoelr
"""
import re
import biosteam as bst
__all__ = ('elementname',)

def elementname(element):
    if element:
        if isinstance(element, type):
            return re.sub(r"\B([A-Z])", r" \1", element.__name__.replace('_', ' ')).capitalize()
        elif isinstance(element, str):
            return element.replace('_', ' ')
        else:
            if isinstance(element, bst.Stream):
                return 'Stream-' + element.ID.replace('_', ' ')
            else:
                return element.line + '-' + element.ID.replace('_', ' ')
    else:
        return 'None'


# -*- coding: utf-8 -*-
"""
Created on Wed Apr 10 11:48:29 2019

@author: yoelr
"""

from ..._unit import metaUnit

__all__ = ('final',)

class final(metaUnit):
    """Create final subclass which is unable to be inherited."""
    def __new__(mcl, name, bases, dct):
        try:
            base, = bases
        except:
            TypeError('cannot create {mcl.__name__} subclass from more than one super class')
            
        if isinstance(base, mcl):
            raise TypeError(f"cannot inherit from {base}. Instances of {mcl.__name__} cannot be inherited")

        return super().__new__(mcl, name, bases, dct)


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:18:36 2018

@author: yoelr
"""
import numpy as np
from .. import Unit, Stream
from .decorators import cost
from .metaclasses import splitter

@cost('Solids loading', cost=68040, CE=567, n=0.50, ub=40, BM=2.03)
class SolidsCentrifuge(Unit, metaclass=splitter):
    """Create a solids centrifuge that separates out solids according to user defined split. Assume a continuous scroll solid bowl. 
    
    **Parameters**
    
        **split:** [array_like] Component splits to 0th output stream
        
        **solids:** tuple[str] IDs of solids 
    
    **ins**
    
        [:] Input streams
    
    **outs**
    
        [0] Liquid stream
        
        [1] Solids stream
    
    **References**
    
        .. [0] Seider, Warren D., et al. (2017). "Cost Accounting and Capital Cost Estimation". In Product and Process Design Principles: Synthesis, Analysis, and Evaluation (pp. 481-485). New York: Wiley.
    
    """
    _kwargs = {'solids': None}
    _units = {'Solids loading': 'tonn/hr'}
    _minimum_solids_loading = 2

    def _init(self):
        self._solids_index = Stream.indices(self._kwargs['solids'])
    
    def _design(self):
        mass_solids = 0
        index = self._solids_index
        for s in self.ins:
            mass_solids += s.mass[index]
        ts = np.asarray(mass_solids).sum() # Total solids
        ts *= 0.0011023 # To short tons (2000 lbs/hr)
        self._Design['Solids loading'] = ts
        lb = self._minimum_solids_loading
        if ts < lb:
            self._lb_warning('Solids loading', ts, 'tonn/hr', lb)
    



# -*- coding: utf-8 -*-
"""
Created on Mon May 20 22:04:02 2019

@author: yoelr
"""

from ... import Stream
from ..._unit import metaUnit
from numpy import asarray
from pandas import Series

__all__ = ('splitter', '__init__', 'run_split', 'run_split_with_mixing')

class splitter(metaUnit):
    """Create a splitter Unit class which behaves like a splitter regarding mass and energy balances.
    
    **Examples**
    
    Create a sugar cane crushing mill Unit subclass and test:
    
    >>> from biosteam import *
    >>> from biosteam.units.decorators import *
    >>> from biosteam.units.metaclasses import splitter
    >>> # Create subclass
    >>> @cost('Flow rate', units='kg/hr', cost=1.5e6,
    ...       CE=541.7, exp=0.6, S=335e3, kW=2010)
    ... class CrushingMill(Unit, metaclass=splitter): pass
    >>> # Set Stream.species
    >>> species = Species('Water')
    >>> species.Sugar = compounds.Substance('Sugar')
    >>> species.Fiber = compounds.Substance('Fiber')
    >>> Stream.species = species
    >>> # Create CrushingMill object and simulate
    >>> CM = CrushingMill(ins=Stream(Water=700, Fiber=150, Sugar=150, units='kg/min'), 
    ...                   split=(0.86, 0.08, 0.96),
    ...                   order=('Water', 'Fiber', 'Sugar'))
    >>> CM.simulate()
    >>> CM.show()
    CrushingMill: U1
    ins...
    [0] d2
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  2.33e+03
                        Sugar  9e+03
                        Fiber  9e+03
    outs...
    [0] d3
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  2e+03
                        Sugar  8.64e+03
                        Fiber  720
    [1] d4
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  326
                        Sugar  360
                        Fiber  8.28e+03
    
    """
    def __new__(mcl, name, bases, dct):
        if '__init__' in dct:
            raise TypeError(f"cannot use {mcl.__name__} metaclass with an implemented '__init__' method")
        elif 'split' in dct:
            raise TypeError("cannot use {mcl.__name__} metaclass with an implemented 'split' member")
        
        _kwargs = dct.get('_kwargs')
        if _kwargs is None:
            dct['__init__'] = __init__
        elif 'split' in _kwargs or 'order' in _kwargs:
            raise TypeError("cannot use {mcl.__name__} metaclass with 'split' or 'order' defined in '{clsname}._kwargs'")
        else:
            # Begin changing __init__ to have kwargs
            inputs = ', '.join([key + '=' + key for key in _kwargs])
            
            # Make a string to execute
            str2exec = f"def __init__(self, ID='', outs=(), ins=None, order=None, {inputs}, *, split):\n"
            str2exec+= f"    self._kwargs = dict({inputs})\n"
            str2exec+= f"    _(self, ID, outs, ins, order, split=split)"
            
            # Execute string and replace __init__
            globs = {'_': __init__}
            globs.update(_kwargs)
            exec(str2exec, globs, dct)
        
        if '_run' not in dct:
            dct['_run'] = run_split
        
        if '_N_ins' not in dct: dct['_N_ins'] = 1
        dct['_N_outs'] = 2
        dct['split'] = split
        
        if "__doc__" in dct:
            dct['__doc__'] = dct['__doc__'].replace('**Parameters**', 
       "**Parameters**"
+"\n"    
+"\n        **split:** Should be one of the following"
+"\n            * [float] The fraction of net feed in the 0th output stream"
+"\n            * [array_like] Componentwise split of feed to 0th output stream"
+"\n"
+"\n        **order:** Iterable[str] Species order of split")    
        
        return super().__new__(mcl, name, bases, dct)

@property
def split(self):
    """[array] Componentwise split of feed to 0th output stream."""
    return self._split_series

def __init__(self, ID='', outs=(), ins=None, order=None, *, split):
    self.ID = ID
    self._split = split = Stream.species.array(order, split) if order else asarray(split)
    self._split_series =  Series(split, Stream.species.IDs)
    self._init_ins(ins)
    self._init_outs(outs)
    self._init_results()
    self._init_heat_utils()
    self._init_power_util()
    self._init()
    
def run_split(self):
    """Splitter mass and energy balance function based on one input stream."""
    top, bot = self.outs
    feed = self._ins[0]
    net_mol = feed.mol
    top._mol[:] = net_mol * self._split
    bot._mol[:] = net_mol - top._mol
    bot.T = top.T = feed.T
    bot.P = top.P = feed.P
    
def run_split_with_mixing(self):
    """Splitter mass and energy balance function with mixing all input streams."""
    top, bot = self._outs
    ins = self._ins
    if len(ins) > 1: Stream.sum(top, ins)
    else: top.copylike(ins[0])
    bot.copylike(top)
    top._mol[:] *= self._split
    bot._mol[:] -= top._mol


# -*- coding: utf-8 -*-
"""
Created on Sun Feb 17 16:36:47 2019

@author: yoelr
"""
from ... import Unit
from ..decorators import cost

__all__ = ('BoilerTurbogenerator',)

#: TODO add reference of NREL

@cost('Heat', CE=567.3, cost=36e6, S=104e7, n=0.95)
class BoilerTurbogenerator(Unit):
    """Create a Boiler_TurboGenerator object that will calculate electricity generation from burning the feed. It also takes into account how much steam is being produced, and the required cooling utility of the turbo generator. No emissions or mass balances are taken into account.
    
    **Parameters**
    
        **boiler_efficiency:** [float] Fraction of heat transfered to steam.
    
        **turbo_generator_efficiency:** [float] Fraction of steam heat converted to electricity.

        **steam:** [Stream] Steam product from the boiler.
    
    **ins**
    
        [0] Feed that will be burned.
    
    **outs**
    
        [0] Emission (burned feed)
        
    
    """
    _N_heat_utilities = 2
    _has_power_utility = True
    _units = {'Heat': 'kJ/hr'}
    _kwargs = {'boiler_efficiency': 0.80,
               'turbo_generator_efficiency': 0.85,
               'steam': None}
    
    def _design(self):
        B_eff, TG_eff, steam = self._kwargs.values()
        feed = self.ins[0]
        emission, = self.outs
        hu_cooling, hu_steam = self._heat_utilities
        lps = hu_steam.heating_agents['Low pressure steam']
        H_steam = steam.H
        self._Design['Heat'] = feed_Hc = feed.Hc
        
        # This is simply for the mass balance (no special purpose)
        emission.mol[:] = feed.mol # TODO: In reality, this should be CO2
        emission.T = steam.T
        emission.phase = 'g'
        
        feed_massnet = feed.massnet
        H_content = feed_Hc*B_eff
        H_electricity = H_content - H_steam - feed_massnet*1130 # 50 percent of bagasse is water, so remove latent heat of vaporization
        if H_electricity < 0:
            H_steam = H_content
            cooling = electricity = 0
        else:
            electricity = H_electricity * TG_eff
            cooling = electricity - H_electricity
        hu_cooling(cooling, steam.T)
        hu_steam.duty = -H_steam
        hu_steam.ID = 'Low pressure steam'
        hu_steam.flow = molnet = -steam.molnet
        hu_steam.cost = molnet*lps.price_kmol + H_steam*lps.price_kJ
        self._power_utility(-electricity/3600)
        

# Simulation of ethanol production from sugarcane
# in Brazil: economic study of an autonomous
# distillery
# Marina O.S. Diasa,b, Marcelo P. Cunhaa, Charles D.F. Jesusa, Mirna I.G.
# Scandiffioa, Carlos E.V. Rossella,b, Rubens Maciel Filhob, Antonio Bonomia
# a CTBE – Bioethanol Science and Technology National Laboratory, PO Box 6170 –
# CEP 13083-970, Campinas – SP, Brazil, marina.dias@bioetanol.org.br
# b School of Chemical Engineering, University of Campinas, PO Box 6066 – CEP
# 13083-970, Campinas – SP, Brazil
# LHV = 7565 # Bagasse lower heating value (kJ/kg)


# -*- coding: utf-8 -*-
"""
Created on Wed Dec  5 17:24:01 2018

This module includes arbitrary classes and functions.

@author: Guest Group
"""
from biosteam import _Q

__all__ = ('factor', 'checkbounds', 'approx2step', 'strtuple', 'function')


# %% Number functions

def factor(base_units, new_units):
    if base_units == new_units: return 1
    else: return _Q(1, base_units).to(new_units).magnitude

def checkbounds(x, bounds):
    lb, up = bounds
    return lb < x < up

def approx2step(val, x0, dx):
    """Approximate value, val, to closest increment/step, dx, starting from x0."""
    while True:
        if x0 > val: break
        x0 += dx
    return x0


# %% String functions

function = type(checkbounds)

def strtuple(iterable):
    """Return string of all items in the tuple""" 
    string = ''
    for i in iterable:
        if isinstance(i , function):
            string += i.__name__ + ', '
        else:
            string += str(i) + ', '
    string = string.rstrip(', ')
    string = '(' + string + ')'
    return string
        




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 28 17:28:09 2017

@author: Yoel Rene Cortes-Pena
"""

__all__ = ['units', 'evaluation', 'inspect', 'find', 'compounds', 'reaction']

CE = 567.5 #: Chemical engineering plant cost index (defaults to 567.5 at 2017)

# %% Import base utils

import pandas as pd
import numpy as np
from pint import UnitRegistry
import os

# Set pint Unit Registry
_ureg = UnitRegistry()
_ureg.default_format = '~P'
_ureg.load_definitions(os.path.dirname(os.path.realpath(__file__)) + '/my_units_defs.txt')
_Q = _ureg.Quantity
_Q._repr_latex_ = _Q._repr_html_ = _Q.__str__ = _Q.__repr__ = lambda self: self.__format__('')

# Set number of digits displayed
np.set_printoptions(suppress=False)
np.set_printoptions(precision=3) 
pd.options.display.float_format = '{:.3g}'.format
pd.set_option('display.max_rows', 35)
pd.set_option('display.max_columns', 10)
pd.set_option('max_colwidth', 35)
del np, pd, os, UnitRegistry

# %% Import biosteam classes

from ._species import *
from ._stream import *
from ._mixed_stream import *
from ._heat_utility import *
from ._power_utility import *
from ._unit import *
from ._system import *
from ._tea import *
from ._flowsheet import *

from . import reaction
from . import inspect
from . import compounds
from . import _species 
from . import _stream
from . import _mixed_stream
from . import _unit
from . import _heat_utility
from . import _power_utility
from . import evaluation
from . import _system 
from . import _tea
from . import _flowsheet
from . import units

__all__.extend(_species.__all__)
__all__.extend(_stream.__all__)
__all__.extend(_mixed_stream.__all__)
__all__.extend(_heat_utility.__all__)
__all__.extend(_power_utility.__all__)
__all__.extend(_unit.__all__)
__all__.extend(_system.__all__)
__all__.extend(_tea.__all__)
__all__.extend(_flowsheet.__all__)



# -*- coding: utf-8 -*-
"""
Created on Sun May 12 17:15:43 2019

@author: yoelr
"""
# def cartesian_samples(parameter_values):
#     tuple_ = tuple
#     initial_sample = []
#     sorted_ = sorted
#     paramvals = []
#     for args in parameter_values:
#         args = sorted_(args)
#         paramvals.append(args)
#         initial_sample.append(args[0])
#     num_args = tuple_(enumerate(paramvals))
#     sample = initial_sample
#     initial_sample = tuple_(initial_sample)
#     samples = [initial_sample]
#     _fillspace(sample, num_args, samples, tuple_)
#     return samples

# def _fillspace(args, num_args, argspace, tuple_):
#     for i, fargs in num_args:
#         for a in fargs[1:]:
#             args[i] = a
#             argspace.append(tuple_(args))
#             _fillspace(args, num_args[:i], argspace, tuple_)
#         fargs.reverse()

# def _split(samples, i):
#     subgrids = {}
#     for s in samples:
#         k = s[i]
#         if k in subgrids: subgrids[k].append(s)
#         else: subgrids[k] = [s]
#     return subgrids.values()

# def _sort(grid, i, end):
#     if i == end: return grid
#     subgrids = _split(grid, i)
#     i_next = i+1
#     key = lambda x: x[i_next]
#     grid.clear()
#     reverse = False
#     for samples in subgrids:
#         samples.sort(key=key, reverse=reverse)
#         reverse = not reverse
#         grid.extend(_sort(samples, i_next, end))
#     return grid
    
# def sort(grid):
#     grid.sort(key=lambda x: x[0])
#     end = len(grid[0]) - 1
#     _sort(grid, 0, end)
#     return grid


# -*- coding: utf-8 -*-
"""
Created on Fri Feb 22 17:31:50 2019

@author: yoelr
"""
import numpy as np
from biosteam._utils import checkbounds
from biosteam._exceptions import DesignError
import biosteam as bst

__all__ = ('vacuum_system', '_calc_MotorEfficiency', '_calc_BreakEfficiency')

ln = np.log

# %% Data

# System types of vacuum systems
# Volumetric flowrate ranges, (cfm) and lower limit of suction (torr)
_steamjet_ejectors = {
    'One stage':               ((10, 1000000), 100),
    'Two stage':               ((10, 1000000),  15),
    'Three stage':             ((10, 1000000),   2)}
_liquid_ring = {
    'One stage water seal':    (( 3,   18000),  50),
    'Two stage water seal':    (( 3,   18000),  25),
    'Oil seal':                (( 3,   18000),  10)}
_dry_vacuum = {
    'Three-stage rotary lobe': ((60,     240), 1.5),
    'Three-stage claw':        ((60,     270), 0.3),
    'Screw compressor':        ((50,    1400), 0.1)}

_default_vacuum_systems = {'Liquid-ring pump': _liquid_ring,
                           'Steam-jet ejector': _steamjet_ejectors,
                           'Dry-vacuum pump': _dry_vacuum}
            
_air_density = 1.2041 # kg/m3 dry air

# %% Calculate vacuum system requirements

def vacuum_system(massflow:'kg/hr', volflow:'m3/hr',
                  P_suction:'Pa', vol:'m3',
                  vacuum_system_preference=None):
    """Return dictionary of results
    
    **Parameters**
    
        **massflow:** [float] Vapor mass flow rate entering vacuum system from vessel (not including inleakage) 
        
        **volflow:** [float] Vapor volumetric flow rate entering vacuum system from vessel (not including inleakage)
        
        **P_suction:** [float] Suction pressure
        
        **vol:** [float] Vacuum volume
        
        **vacuum_system_preference:** tuple[str] or [str] Name(s) of preferred vacuum systems
    
    **Return**
    
        **power:** [float] (kW)
        
        **cost:** [float] (USD)
    
    """
    vacuum_systems = _prefered_VacuumSystems(vacuum_system_preference)
    P_suction *= 7.5006e-3 # to torr
    if vol:
        vol *= 35.315 # to ft3
        air_massflow = _calc_AirInleakage(vol, P_suction) # lb/hr
        air_volflow = 0.26697*air_massflow/_air_density # cfm
    else:
        air_massflow = 0
        air_volflow = 0
    volflow_cfm = 0.5886*volflow + air_volflow
    if volflow_cfm < 3.01:
        factor = 3.01/volflow_cfm
        volflow_cfm = 3.01
    else:
        factor = 1
    massflow_kgph = (massflow + 0.4536*air_massflow)*factor # kg/hr
    massflow_lbph = 2.205*massflow_kgph
    power = _power(massflow_kgph, P_suction)
    vacuum_sys, grade = _select_VacuumSystem(vacuum_systems, volflow_cfm, P_suction)
    cost = bst.CE*_cost(vacuum_sys, grade, massflow_lbph, volflow_cfm, P_suction)/567
    return power, cost

# %% Supporting functions

def _calc_BreakEfficiency(q:'gpm'):
    if q < 50: q = 50
    elif q > 5000: q = 5000
    return -0.316 + 0.24015*ln(q) - 0.01199*ln(q)**2

def _calc_MotorEfficiency(Pb):
    if Pb < 1: Pb = 1
    elif Pb > 1500: Pb = 1500
    return 0.8 + 0.0319*ln(Pb) - 0.00182*ln(Pb)**2

def _prefered_VacuumSystems(preference):
    if preference is None:
        vacuum_systems = _default_vacuum_systems.values()
    else:
        defaults = _default_vacuum_systems.keys()
        if isinstance(preference, str):
            if preference not in defaults:
                raise ValueError(f"preference have at least one of the following: {defaults}")
            preference = (preference,)
        else:
            for name in preference:
                if name not in defaults:
                    raise ValueError(f"preference have at least one of the following: {defaults}")
        
        vacuum_systems = [_default_vacuum_systems[name] for name in preference]
    return vacuum_systems


def _available_VacuumSystems(flow:'cfm', P_suction:'torr'):
    types = []
    for vacuumtype, vacuum_sys in _default_vacuum_systems.items():
        for grade, flowrange_minsuction in vacuum_sys.items():
            flowrange, minsuction = flowrange_minsuction
            if checkbounds(flow, flowrange) and P_suction > minsuction:
                types.append((vacuumtype, grade))
    return types

def _select_VacuumSystem(vacuum_systems, volflow, P_suction):
    for vacuum_sys in vacuum_systems:
        for grade, flowrange_minsuction in vacuum_sys.items():
            flowrange, minsuction = flowrange_minsuction
            if checkbounds(volflow, flowrange) and P_suction > minsuction:
                return (vacuum_sys, grade)
    raise DesignError('no vacuum system available at current flow and suction pressure')

def _lazycalc_AirInleakage(V:'ft^3', P:'Pa') -> 'lb/hr':
    if P > 11999.013: k = 0.2
    elif P > 4132.993: k = 0.15
    elif P > 2799.77: k = 0.10
    elif P > 133.322: k = 0.051
    else: raise ValueError('cannot calculate air inleakage at pressures lower than 133.32 Pascal')
    return k*V**0.667

def _calc_AirInleakage(V:'ft^3', P:'torr') -> 'lb/hr':
    return 5 + (0.0298 + 0.03088*ln(P) - 5.733e-4*ln(P)**2)*V**0.66
    
def _power(massflow:'kg/hr',  P_suction:'torr') -> 'kW':
    SF = massflow/P_suction # Size factor
    if SF < 0.2: SF = 0.2
    elif SF > 16: SF = 16
    Pb = 21.4*SF**0.924 # Break power assuming Liquid-ring (NASH)
    mu_m = _calc_MotorEfficiency(Pb)
    return Pb/mu_m

def _cost(vacuum_sys, grade, massflow:'lb/hr', volflow:'cfm', P_suction:'torr') -> 'USD':
    # Costing for different vacuums ASSUMING THAT ALL ARE CARBON STEEL!!!!
    if vacuum_sys is _steamjet_ejectors:
        S = massflow/P_suction
        Cp = 1915*(S**(0.41))
        if grade == 'One stage':
            Cs = 1
        elif grade == 'Two stage':
            Cs = 1.8
        elif grade == 'Three stage':
            Cs = 2.1
        Cost = Cp*Cs
    elif vacuum_sys is _liquid_ring:
        S = volflow
        Cp = 8250*(S**(0.37))
        if grade == 'One stage water seal':
            Cs = 1
        elif grade == 'Two stage water seal':
            Cs = 1.8
        elif grade == 'Oil seal':
            Cs = 2.1
        Cost = Cp*Cs
    elif vacuum_sys is _dry_vacuum:
        S = volflow
        if grade == 'Three-stage rotary lobe':
            Cp = 8075*(S**(0.41))
        elif grade == 'Three-stage claw':
            Cp = 9785*(S**(0.36))
        elif grade == 'Screw compressor':
            Cp = 10875*(S**(0.38))
        Cost = Cp
    return Cost
    



# -*- coding: utf-8 -*-
"""
Created on Wed Dec  5 16:47:33 2018

@author: Yoel Cortes-Pena
"""
from weakref import ref

__all__ = ('Register', 'search_register')

getattr_ = object.__getattribute__
setattr_ = object.__setattr__

def _reload(self):
    dict_ = getattr_(self, '__dict__')
    # Dictionary will change size in for loop, so use list
    for i, r in list(dict_.items()):
        j = r()
        if not j:
            del dict_[i]
        else:
            s = str(j)
            if i!=s: del dict_[i]
            dict_[s] = r
    _must_reload.remove(self)

def search_register(self, key):
    ref =  getattr_(self, '__dict__').get(key)
    return ref() if ref else ref 

_must_reload = set()

class Register:
    
    def __getattribute__(self, key):
        if self in _must_reload: _reload(self)
        try: return getattr_(self, key)()
        except: return getattr_(self, key)
    __getitem__ = __getattribute__
    
    def __bool__(self):
        return bool(getattr_(self, '__dict__'))
    
    def __setattr__(self, key, value):
        setattr_(self, key, ref(value))
        _must_reload.add(self)
    __setitem__ = __setattr__
    
    def __iter__(self):
        if self in _must_reload: _reload(self)
        for i in getattr_(self, '__dict__').values(): yield i()
    
    def __delattr__(self, key):
        dict_ = getattr_(self, '__dict__')
        value = dict_[key]()
        if hasattr(value, '_disconnect'): value._disconnect()
        del dict_[key]
    __delitem__ = __delattr__
    
    def __repr__(self):
        if getattr_(self, '__dict__'):
            if self in _must_reload: _reload(self)
            return f'Register:\n ' + '\n '.join([repr(i) for i in self])
        else:
            return f'Register: (Empty)'




# -*- coding: utf-8 -*-
"""
Created on Sat Jun 15 00:46:41 2019

@author: yoelr
"""
from .. import Unit, units
from . import metaclasses
from . import decorators
import pandas as pd
import numpy as np

_add_cost = decorators.add_cost
_index = np.array(['Basis',
                   'Units',
                   'Size',
                   'Upper bound',
                   'CEPCI',
                   'Cost (USD)',
                   'Exponent',
                   'Electricity (kW)',
                   'Installation factor'])


__all__ = ('df2unit', 'xl2dct', 'xl2mod')

def df2unit(clsname, cost_items, *, supercls=None, metacls=None):
    """Return Unit subclass from cost_options DataFrame."""
    superclasses = (supercls,) if supercls else (Unit,)
    if not metacls: metacls = type(supercls)
    try: assert all(cost_items.index==_index), "'cost_items' index is incorrect"
    except Exception as err:
        if not isinstance(cost_items, pd.DataFrame):
            raise ValueError(f"'cost_items' must be a DataFrame, not '{type(cost_items).__name__}'")
        else:
            raise err
    cls = metacls.__new__(metacls, clsname, superclasses, {})
    for ID in cost_items:
        _add_cost(cls, ID, *cost_items[ID], None)
    return cls

def df2dct(df):
    dct = {}
    for name_sim in df.columns.levels[0]:
        cost_items = df[name_sim]
        if '-' in name_sim:
            sim, name = name_sim.split('-')
        else:
            name = name_sim
            sim = 'static'
        name = ''.join([i.capitalize() for i in name.split(' ')])
        for i in name:
            if not (i is ' ' or i.isalnum()):
                name = name.replace(i, ' ')
        metaname = sim.casefold()
        if metaname in metaclasses.__dict__:
            metacls = getattr(metaclasses, metaname)
            new = df2unit(name, cost_items, metacls=metacls)
        else:
            supername = ''.join([i.capitalize() for i in sim.split(' ')])
            if supername in units.__dict__:
                superclass = getattr(units, supername)
            else:
                raise ValueError(f"invalid simulation option '{sim}'")
            new = df2unit(name, cost_items, superclass=superclass)
        dct[name] = new
    return dct

def xl2dct(file, sheet_name=0):
    """Return dictionary of unit subclasses from excel file."""
    return df2dct(pd.read_excel(file, header=[0, 1]))

def xl2mod(file, module, sheet_name=0):
    dct = xl2dct(file, sheet_name)
    for i, j in dct.items():
        setattr(module, i, j)
        j.__module__ = module.__name__
    if not hasattr(module, '__all__'):
        module.__all__ = tuple(dct)
    
    
    
    
    
    


# -*- coding: utf-8 -*-
"""
Created on Sat Jun  8 23:29:37 2019

@author: yoelr
"""
from .._unit import Unit
from .metaclasses import final
from .._stream import Stream
from .._mixed_stream import MixedStream
from .._utils import missing_stream, Ins, Outs
from .._flowsheet import find

__all__ = ('Junction',)

def _getstreams(self):
    try:
        upstream, = self._ins
        downstream, = self._outs
    except ValueError as error:
        N_ins = len(self._ins)
        N_outs = len(self._outs)
        if N_ins != 1:
            raise RuntimeError(f'a Junction object must have 1 input stream, not {N_ins}')
        elif N_outs != 1:
            raise RuntimeError(f'a Junction object must have 1 output stream, not {N_outs}')
        else:
            raise error
    return upstream, downstream

# %% Connect between different flowsheets

default_species = lambda upstream, downstream: \
        set(upstream.species._IDs).intersection(downstream.species._IDs)

class Junction(Unit, metaclass=final):
    """Create a Junction object that copies specifications from `upstream`
    to `downstream`. This serves to connect streams with different
    Species object.
    
    **Parameters**
        
        **downstream:** [Stream or str] Flow rate, T, P, and phase information
        will be copied from `upstream` to this stream.
        If None, stream will be missing.
        
        **upstream:** [Stream or str] Stream that will be copied to
        `downstream`. If None, stream will be missing.
        
        **species:** list[str] IDs of species to be passed down.
        If None, all species in common will be passed.
    
    **Examples**
    
    Create a Junction object and connect streams with different Species objects:
        
    >>> from biosteam import *
    >>> species = Species('Water')
    >>> species.Sugar = compounds.Substance('Sugar')
    >>> Stream.species = species
    >>> s1 = Stream('s1', Water=20)
    >>> Stream.species = Species('Water', 'Ethanol')
    >>> s2 = Stream('s2') # Note that s2 and s1 have different Species objects
    >>> J1 = units.Junction(s1, s2)
    >>> J1.simulate()
    >>> J1.show()
    Junction:
    ins...
    [0] s1
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  20
    outs...
    [0] s2
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  20
        
    """
    _has_cost = False
    _N_ins = _N_outs = 1
    def __init__(self, upstream=None, downstream='', species=None):
        # Init upstream
        if upstream is None:
            self._ins = Ins(self, (missing_stream,))
        elif isinstance(upstream, Stream):
            self._ins = Ins(self, (upstream,))
        elif isinstance(upstream, str):
            self._ins = Ins(self, (Stream(upstream),))
        
        # Init downstream
        if downstream is None:
            self._outs = Outs(self, (missing_stream,))
        elif isinstance(downstream, Stream):
            self._outs = Outs(self, (downstream,))
        elif isinstance(downstream, str):
            self._outs = Outs(self, (Stream(downstream),))
        
        find.unit[self.ID] = self
        upstream, downstream = _getstreams(self)
        if upstream and downstream:
            self.species = default_species(upstream, downstream)
        else:
            self._species = None
    
    def _run(self):
        upstream, downstream = _getstreams(self)
        try:
            if isinstance(upstream, MixedStream):
                downstream.enable_phases()
                downstream._mol[:, self._downindex] = upstream._mol[:, self._upindex]
            else:
                downstream.disable_phases(upstream._phase)
                downstream._mol[self._downindex] = upstream._mol[self._upindex]
        except AttributeError as error:
            species = default_species(upstream, downstream)
            if self._species and all([(i in self._species) for i in species]): raise error
            downstream._mol[:] = 0
            self.species = species
            self._run()
        downstream.T = upstream.T
        downstream.P = upstream.P
    simulate = _run
    
    @property
    def ID(self):
        upstream, downstream = _getstreams(self)
        if (upstream or downstream):
            return f"{self._ins[0]} to {self._outs[0]}"
        else:
            return "missing streams"
    _ID = ID
    
    @property
    def species(self):
        return self._species
    @species.setter
    def species(self, IDs):
        upstream, downstream = _getstreams(self)
        self._species = tuple(IDs)
        self._upindex = upstream.indices(IDs)
        self._downindex = downstream.indices(IDs)
    
    def _info(self, T, P, flow, fraction):
        info = super()._info(T, P, flow, fraction)
        return info[:info.index(':')+1] + info[info.index('\n'):]
    
    def __repr__(self):
        return f"<{type(self).__name__}: {self.ID}>"


def node_function(self):
    node = self._graphics.node
    if not any(_getstreams(self)):
        node['fontsize'] = '18'
        node['shape'] = 'plaintext'
        node['fillcolor'] = 'none'
    else:
        node['width'] = '0.1'
        node['shape'] = 'point'
        node['color'] = node['fillcolor'] = 'black'

Junction._graphics.node_function = node_function
del node_function



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:14:01 2018

@author: yoelr
"""
from .. import Unit
from .._exceptions import UndefinedCompound
from .decorators import cost
from .metaclasses import splitter, run_split_with_mixing
import biosteam as bst

@cost('Flow rate', units='kg/hr', cost=1.5e6, CE=541.7,
      n=0.6, S=335e3, kW=2010, BM=2.3)
class CrushingMill(Unit, metaclass=splitter):
    """Create CrushingMill object.
    
    **Parameters**
    
        **moisture_content:** Fraction of water in Baggasse
    
    **ins**
    
        [0] Shredded sugar cane
        
        [1] Recycle water
    
    **outs**
    
        [0] Bagasse
        
        [1] Juice
    
    """
    _N_ins = 2
    _kwargs = {'moisture_content': 0.5}
    def _init(self):
        try:
            self._water_index = wi = bst.Stream.species._indexdct['7732-18-5']
        except KeyError:
            raise UndefinedCompound('7732-18-5')
        if self._split[wi] != 0:
            raise ValueError('cannot define water split, only moisture content')
    
    def _run(self):
        run_split_with_mixing(self)
        wi = self._water_index
        bagasse = self.outs[0]
        juice = self.outs[1]
        solids = bagasse.massnet
        mc = self._kwargs['moisture_content']
        bagasse._mol[wi] = water = (solids * mc/(1-mc))/18.01528
        juice._mol[wi] -= water
        if juice._mol[wi] < water:
            raise ValueError(f'not enough water for {repr(self)}')
    



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 14:26:41 2018

@author: yoelr
"""
from ... import Stream
from ..._unit import metaUnit

__all__ = ('mixer', 'run_mixer')

class mixer(metaUnit):
    """Create a mixer Unit class which behaves like a mixer regarding mass and energy balances."""
    def __new__(mcl, name, bases, dct):
        if '_run' in dct:
            raise TypeError(f"cannot use {mcl.__name__} metaclass with an implemented '_run' method")
        dct['_run'] = run_mixer
        dct['_N_ins'] = 2
        dct['_N_outs'] = 1
        return super().__new__(mcl, name, bases, dct)

run_mixer = lambda self: Stream.sum(self.outs[0], self.ins)
    
    



# -*- coding: utf-8 -*-
"""
Created on Mon Apr 29 18:28:55 2019

@author: yoelr
"""
from ... import Unit
from ..decorators import cost
from ..metaclasses import static
from copy import copy

__all__ = ('CoolingTower', 'CoolingTowerWithPowerDemand')

@cost('Flow rate', 'Cooling tower', cost=1100, CE=567, n=0.7)
class CoolingTower(Unit, metaclass=static):
    """Create a cooling tower that is cost based on flow rate of cooling water.
    
    **ins**
    
        [0] Cooling water
        
    **outs**
    
        [0] Cooling water
    
    """
    _units = {'Flow rate': 'gpm'}
    def _design(self):
        self._Design['Flow rate'] = self._ins[0].molnet*0.07932

   
class CoolingTowerWithPowerDemand(CoolingTower):
    _has_power_utility = True
    _N_heat_utilities = 1
    cost_options = copy(CoolingTower.cost_items)
    cost_options['Cooling tower'].kW = 0.25215
    def _cost(self):
        super()._cost()
        q = self.ins[0].molnet # kmol/hr
        hu = self._heat_utilities[0]
        cw = hu.cooling_agents['Cooling water']
        hu.ID = 'Cooling water'
        hu.flow = -q
        hu.cost = -q*cw.price_kmol


# -*- coding: utf-8 -*-
"""
Created on Sat Jun 29 20:47:24 2019

@author: yoelr
"""

__all__ = ('str2dct', 'dct2str', 'dct2arr', 'arr2dct',  'str2arr', 'arr2str')

def str2dct(reaction) -> dict:
    reaction = reaction.replace(' ', '')
    left, right = reaction.split('->')
    reactants = left.split('+')
    products = right.split('+')
    dct = {}
    for nID in reactants:
        for i, letter in enumerate(nID):
            if letter.isalpha(): break
        if i: dct[nID[i:]] = -float(nID[:i])
        else: dct[nID] = -1
    for nID in products:
        for i, letter in enumerate(nID):
            if letter.isalpha(): break
        if i: dct[nID[i:]] = float(nID[:i])
        else: dct[nID] = 1
    return dct

def dct2str(dct):
    left = []
    right = []
    for ID, N in dct.items():
        N_int = int(N)
        if N_int == N: N = N_int
        if N == -1: left.append(ID)
        elif N == 1: right.append(ID)
        elif N < 0: left.append(f"{-N} {ID}")
        else: right.append(f"{N} {ID}")
    left = ' + '.join(left)
    right = ' + '.join(right)
    reaction = left + ' -> ' + right
    return reaction

def dct2arr(dct, species):
    return species.array(dct.keys(), [*dct.values()])

def arr2dct(arr, species):
    dct = {}
    for N, ID in zip(arr, species._IDs):
        if N: dct[ID] = N
    return dct

def str2arr(reaction, species):
    return dct2arr(str2dct(reaction), species)

def arr2str(arr, species):
    return dct2str(arr2dct(arr, species))



# -*- coding: utf-8 -*-
"""
Created on Fri Jun  7 00:53:29 2019

@author: yoelr
"""
from ... import Unit

__all__ = ('extend_summary', )

def extend_summary(cls):
    """Extends the Unit class with the following abstract methods:
        
    **_end():**
        Finish setting purchase prices and utility costs.
        
    """
    if hasattr(cls, '_end'):    
        if cls._summary is Unit._summary:
            cls._summary = _summary
        elif cls._summary is not _summary:
            raise RuntimeError("cannot decorate Unit subclass an implemented '_summary' method")

def _summary(self):
    """Calculate all results from unit run."""
    self._design()
    self._cost()
    self._end()


# -*- coding: utf-8 -*-
"""
Created on Sun Jun 30 21:21:45 2019

@author: yoelr
"""
from .. import Unit
from .. reaction import Rxn, PRxn

class LignocellulosicPretreatment(Unit):
    
    
    def _init(self):
        self.conversion = PRxn([
            #   Reaction definition                 Reactant    Conversion
            Rxn('Glucan + H2O -> Glucose',          'Glucan',   0.0990),
            Rxn('Glucan + H2O -> Glucose Oligomer', 'Glucan',   0.0030),
            Rxn('Glucan -> HMF + 2 H2O',            'Glucan',   0.0030),
            Rxn('Sucrose -> HMF + Glucose + 2H2O',  'Sucrose',  0.0030),
            Rxn('Xylan + H2O -> Xylose',            'Xylan',    0.9000),
            Rxn('Xylan + H2O -> XyloseOligomer',    'Xylan',    0.0024),
            Rxn('Xylan -> Furfural + 2 H2O',        'Xylan',    0.0050),
            Rxn('Acetate -> AceticAcid',            'Acetate',  1.0000),
            Rxn('Lignin -> SolubleLignin',          'Lignin',   0.0050)])
        
    def _run(self):
        feed = self.ins[0]
        product = self.outs[0]
        prodmol = product.mol
        prodmol[:] = feed.mol
        prodmol += self.conversion(product.mol)
        
        


# -*- coding: utf-8 -*-
"""
Created on Wed Dec  5 17:19:33 2018

This module includes classes and functions relating string coloring and format.

@author: Guest Group
"""
from colorpalette import Color, Palette

__all__ = ('CS', 'color_scheme')

# %% Classes for coloring

ansicolor = Color.ansicolor
CS = color_scheme = Palette(
        exception=ansicolor('\x1b[31m\x1b[1m', 'brightred'),
        info=ansicolor('\x1b[34m\x1b[1m', 'brightblue'),
        note=ansicolor('\x1b[36m\x1b[1m', 'brightcyan'))

CS.dark_warning = Color('darkgoldenrod')
CS.warning = Color('gold')
CS.request = Color('peachpuff')
CS.dim = ansicolor('\x1b[37m\x1b[22m', ID='dim')
CS.reset = ansicolor('\x1b[0m', ID='reset')
CS.ul = Color(style='underline')
CS.num = Color('rednum', '#f74f4f')
CS.str = Color('greenstr', '#89e189')
CS.next = CS.info
CS.bold = ansicolor('\033[1m', ID='bold')
CS.italic = ansicolor('\x1b[3m', ID='italic')


# -*- coding: utf-8 -*-
"""
Created on Wed May  1 19:05:53 2019

@author: yoelr
"""
from inspect import signature
from functools import reduce
from ._extend import extend_finalize

__all__ = ('spec',)

# %% Tracking cost factors
_mul = lambda x, y: x*y

def _spec(self):
    D = self._results['Design']
    C = self._results['Cost']
    S = self._specdata
    for i in S: C[i] *= reduce(_mul, (f(D[p]) for f, p in S[i]))

def spec(item, param, func):
    if not isinstance(param, str):
        raise ValueError(f"param must be a string, not a '{type(param).__name__}' object.")
    N = len(signature(func).parameters)
    if N != 1:
        raise ValueError(f"one and only one argument in 'func' signature is allowed ({N or 'no'} parameters given)")
    
    def spec_decorator(cls):
        extend_finalize(cls)
        if hasattr(cls, '_specdata'): data = cls._specdata
        else: cls._specdata = data = {}
        if item in data: data[item].append((func, param))
        else: data[item] = [(func, param)]
        cls._spec = _spec
        return cls
    return spec_decorator
        
    
    


# -*- coding: utf-8 -*-
"""
Created on Mon Feb  4 19:38:37 2019

@author: Guest Group
"""

import pandas as pd
import numpy as np
from scipy.optimize import newton
from copy import copy

__all__ = ('TEA',)

_DataFrame = pd.DataFrame
_array = np.array

# TODO: Add 'SL', 'DB', 'DDB', 'SYD', 'ACRS' and 'MACRS' functions to generate depreciation data

# %% Depreciation data

_MACRS = {'MACRS5':  _array([.2000, .3200, .1920,
                             .1152, .1152, .0576]),
          
          'MACRS7':  _array([.1429, .2449, .1749,
                             .1249, .0893, .0892,
                             .0893, .0446]),
          
          'MACRS10': _array([.1000, .1800, .1440,
                             .1152, .0922, .0737,
                             .0655, .0655, .0656,
                             .0655, .0328]),

          'MACRS15': _array([.0500, .0950, .0855,
                             .0770, .0693, .0623,
                             .0590, .0590, .0591,
                             .0590, .0591, .0590,
                             .0591, .0590, .0591,
                             .0295])}


# %% Cash flow and results info

_cashflow_columns = ('Depreciable capital',
                     'Depreciation',
                     'Fixed capital',
                     'Working capital',
                     'Annual operating cost (excl. depr.)',
                     'Sales',
                     'Net earnings',
                     'Cash flow',
                     'Discounted cash flow',
                     'Cumulative cash flow')


# %% Techno-Economic Analysis

class TEA:
    """Create a TEA object that can perform cash flow analysis on a System object.
    
        **Parameters**
        
            **system:** [System] Should contain feed and product streams.
            
        **Examples**
        
            :doc:`Techno-economic analysis of a biorefinery` 
    
    """
    __slots__ = ('system', '_cached', '_cashflow',
                 '_options', '_IRR_guess', '_cost_guess',
                 '_costarr', '_units')
    
    #: Default cash flow options
    _default = {'Lang factor': 3,
                'Operating days': 330,
                'IRR': 0.15,
                'Wage': 5e4,
                'Year': None,
                'Employees': 50,
                'Fringe benefits': 0.40,
                'Income tax': 0.35,
                'Property tax': 0.001,
                'Property insurance': 0.005,
                'Duration': 20,
                'Supplies': 0.20,
                'Maintenance': 0.01,
                'Administration': 0.005,
                'Working capital': 0.05,
                'Startup': 0,
                'Land': 0,
                'Royalties': 0,
                'Contingency': 0.18,
                'Depreciation': 'MACRS7',
                'Startup schedule': (0.4, 0.6),
                'Other recurring costs': 0,
                'Other fixed capital': 0}
    
    @property
    def options(self):
        """
        [dict] Options for cash flow analysis [-]:
            * **Lang factor:** Lang factor for getting fixed capital investment from total purchase cost. If no lang factor, estimate capital investment using bare module factors.
            * **Operating days:** (day).
            * **IRR:** Internal rate of return (fraction).
            * **Wage:** Wage per employee (USD/yr).
            * **Year:** Start year of venture.
            * **Employees:** Number of employees.
            * **Fringe benefits:** Cost of fringe benefits as a fraction of labor cost.
            * **Income tax:** Combined federal and state income tax rate (fraction).
            * **Property tax:** Fee as a fraction of fixed capital investment.
            * **Property insurance:** Fee as a fraction of fixed capital investment.
            * **Duration:** Duration of venture (years).
            * **Supplies:** Yearly fee as a fraction of labor costs.
            * **Maintenance:** Yearly fee as a fraction of fixed capital investment.
            * **Administration:** Yearly fee as a fraction of fixed capital investment.
            * **Working capital:** Fee as a fraction of fixed capital investment.
            * **Startup:** Cost of start up as a fraction of depreciable capital.
            * **Land:** Cost of land as a fraction of depreciable capital.
            * **Royalties:** Cost of royalties as a fraction of depreciable capital.
            * **Depreciation:** 'MACRS' + number of years (e.g. 'MACRS7').
            * **Startup schedule:** tuple of startup investment fractions per year, starting from year 0. For example, for 50% capital investment in year 0 and 50% investment in year 1: (0.5, 0.5).
            * **Contingency:** Cost as a fraction of direct permanent invesment.
            * **Other recurring costs:** Any additional recurring costs per year of operation.
            * **Other fixed capital:** Any additional investment not accounted for.
        """
        return self._options
    
    def __init__(self, system):
        self.system = system
        self._options = dict(**self._default)
        
        # [dict] Cached values from '_init' methods:
        #  {operating_days: flowrate_factor,
        #   year_duration: (duration_array, cashflow_data),
        #   depreciation_schedule: (start, Depreciation, D_len, schedule)}
        self._cached = {}
        
        #: Guess IRR for solve_IRR method
        self._IRR_guess = 0.15
        
        #: Guess stream cost for solve_price method
        self._cost_guess = 0
        
        units = system._costunits
        
        #: list[Unit] All unit operations considered
        self._units = units = sorted(units, key=lambda x: x.line)
        self._costarr = np.zeros([len(units), 2])
        
        system._TEA = self

    @property
    def utility_cost(self):
        """Total utility cost (USD/hr)."""
        return sum([u.utility_cost for u in self._units])
    @property
    def purchase_cost(self):
        """Total purchase cost (USD)."""
        return sum([u.purchase_cost for u in self._units])
    @property
    def intallation_cost(self):
        """Total installation cost (USD)."""
        return sum([u.installation_cost for u in self._units])

    def get_cashflow(self):
        """Return DataFrame of the cash flow analysis."""
        flow_factor, cashflow_info, depreciation_data = self._get_cached_data()
        cashflow_data, duration_array = cashflow_info
        parameters = self._calc_parameters(flow_factor)
        self._calc_cashflow(cashflow_data,
                            parameters[:-3],
                            depreciation_data)
        CF, DCF, CPV = cashflow_data[-3:]
        DCF[:] = CF/(1+self.options['IRR'])**duration_array
        CPV[:] = DCF.cumsum()
        return copy(self._cashflow)

    def _results(self):
        """Return a dictionary of summarized results of cash flow analysis."""
        flow_factor, cashflow_info, depreciation_data = self._get_cached_data()
        cashflow_data, duration_array = cashflow_info
        parameters = self._calc_parameters(flow_factor)
        self._calc_cashflow(cashflow_data,
                            parameters[:-3],
                            depreciation_data)
        NPV = self._calc_NPV(self.options['IRR'],
                             cashflow_data[-3],
                             duration_array)
        DC, FCI, WC, S, C, tax, UC, MC, L = parameters
        D = FCI/self._options['Duration']
        AOC = C + D
        TCI = FCI + WC
        net_earnings = (1-tax)*(S-AOC)
        ROI = net_earnings/TCI
        PBP = FCI/(net_earnings + D)
        r = {}
        r['Depreciable capital'] = DC
        r['Fixed capital investment'] = FCI
        r['Working capital'] = WC
        r['Total capital investment'] = TCI
        r['Depreciation'] = D
        r['Utility cost'] = UC
        r['Material cost'] = MC
        r['Sales'] = S
        r['Labor'] = L
        r['Annual operating cost'] = AOC
        r['Net present value'] = NPV
        r['Return on investment'] = ROI
        r['Pay back period'] = PBP
        return r

    def results(self, with_units=True):
        """Return results of techno-economic analysis as a DataFrame object if `with_units` is True or as a Series otherwise."""
        r = self._results()
        keys = []; addkey = keys.append
        vals = []; addval = vals.append
        if with_units:
            results_units = {'Pay back period': 'yr',
                             'Return on investment': '1/yr',
                             'Net present value': 'USD',
                             'Depreciable capital': 'USD',
                             'Fixed capital investment': 'USD',
                             'Total capital investment': 'USD',
                             'Depreciation': 'USD/yr',
                             'Annual operating cost': 'USD/yr',
                             'Working capital': 'USD',
                             'Utility cost': 'USD/yr',
                             'Material cost': 'USD/yr',
                             'Sales': 'USD/yr',
                             'Labor': 'USD/yr'}
            for ki, vi in r.items():
                addkey(ki)
                addval((results_units.get(ki, ''), vi))
            return pd.DataFrame(vals, keys, ('Units', 'Value'))
        else:
            return pd.Series(r)

    def NPV(self):
        """Calculate NPV by cash flow analysis."""
        flow_factor, cashflow_info, depreciation_data = self._get_cached_data()
        cashflow_data, duration_array = cashflow_info
        parameters = self._calc_parameters(flow_factor)
        self._calc_cashflow(cashflow_data,
                            parameters[:-3],
                            depreciation_data)
        return self._calc_NPV(self.options['IRR'],
                              cashflow_data[-3],
                              duration_array)
    
    def production_cost(self, *products):
        """Return production cost of products.
        
        **Parameters**
        
            ***products:** [Stream] Main products of the system
        
        .. Note::
           If there is more than one main product, The production cost is proportionally allocated to each of the main products with respect to their marketing values. The marketing value of each product is determined by the annual production multiplied by its selling price.
        """
        system = self.system
        sysfeeds = system.feeds
        sysproducts = system.products
        o = self.options
        flow_factor = 24*o['Operating days']
        # DPI_: Direct permanent investment (USD)
        # DC_: Depreciable capital (USD)
        # UC_: utility cost (USD/hr)
        costarr = self._costarr
        F_lang = o['Lang factor']
        if F_lang:
            costarr[:] = [(u.purchase_cost, u.utility_cost) for u in self._units]
            DC_, UC_ = costarr.sum(0) 
            DC_ *=  F_lang
        else:
            costarr[:] = [(u.installation_cost, u.utility_cost) for u in self._units]
            DPI_, UC_ = costarr.sum(0) 
            DC_ = DPI_ * (1 + o['Contingency'])
        FC_ = DC_ * (1 + o['Startup'] + o['Land'] + o['Royalties']) + o['Other fixed capital']
        MC_ = 0 # Material cost USD/hr
        CP_ = 0 # Coproducts USD/hr
        for s in sysfeeds:
            price = s.price
            if price: MC_ += price*s.massnet
        for s in sysproducts:
            if s not in products:
                price = s.price
                if price: CP_ += price*s.massnet
        # Multiply by flow_factor for USD/yr
        UC_ *= flow_factor
        MC_ *= flow_factor
        CP_ *= flow_factor
        fb = o['Fringe benefits'] + o['Supplies']
        f =  (o['Maintenance']
            + o['Administration']
            + o['Property tax']
            + o['Property insurance'])
        d = DC_/o['Duration'] # Depreciation
        L_ = o['Wage']*o['Employees']
        total_operating_cost = UC_ + (1+fb)*L_ + MC_ + f*FC_ + d + o['Other recurring costs']
        market_values = np.array([i.cost for i in products])
        weights = market_values/market_values.sum()
        production_cost = weights*total_operating_cost
        return production_cost
    
    def _get_cached_data(self):
        """Return cached data.
        
        **Return**
        
            **flow_factor:** [float] Factor to convert flow rate from kg/hr to kg/yr of operation.
            
            **cashflow_info:** tuple[array] including:
                * cashflow_data: Cash flow data table
                * duration_array: Range from 1 to the end of the project length
        
            **depreciation_data:** [tuple] including:
                * start: Index of year when operation starts
                * Depreciation: Array of depreciation as a fraction of fixed cost
                * D_len: Lenght of depreciation
                * index_startup: Year index and fixed capital investment fractions
            
        """
        # Keys for cached data
        o = self._options
        cached = self._cached
        
        # Get and update cached data
        operating_days = o['Operating days']
        flow_factor = cached.get(operating_days)
        if not flow_factor:
            cached[operating_days] = flow_factor = 24*operating_days
        
        year_duration = (o['Year'], o['Duration'])
        cashflow_info = cached.get(year_duration)
        if not cashflow_info:
            year, duration = year_duration
            index = tuple(range(year, year + duration)) if year else None
            data = np.zeros((duration, 10))
            cashflow_data = data.transpose()
            self._cashflow = _DataFrame(data, index, _cashflow_columns, dtype=float)
            duration_array = _array(range(duration))
            cached[year_duration] = cashflow_info = (cashflow_data, duration_array)
        
        depreciation_schedule = (o['Depreciation'], o['Startup schedule'])
        depreciation_data = cached.get(depreciation_schedule)
        if not depreciation_data:
            depreciation, schedule = depreciation_schedule
            Depreciation = _MACRS[depreciation]
            start = len(schedule)
            end = start + len(Depreciation)
            schedule = np.array(schedule)
            depreciation_data = (start,
                                 Depreciation,
                                 end,
                                 schedule)
            cached[depreciation_schedule] = depreciation_data
        
        return flow_factor, cashflow_info, depreciation_data
    
    def _calc_parameters(self, flow_factor):
        """Return elementary cash flow parameters."""
        # Cash flow parameters (USD or USD/yr)
        # DC_: Depreciable capital
        # FC_: Fixed capital
        # WC_: Working capital
        # UC_: Utility cost
        # MC_: Material cost
        # S_: Sales
        # L_: Labor cost
        # C_: Annual operating cost (excluding depreciation)
        
        system = self.system
        feeds = system.feeds
        products = system.products
        o = self.options
        # DC_: Depreciable capital (USD)
        # UC_: utility cost (USD/hr)
        costarr = self._costarr
        F_lang = o['Lang factor']
        if F_lang:
            costarr[:] = [(u.purchase_cost, u.utility_cost) for u in self._units]
            DC_, UC_ = costarr.sum(0) 
            DC_ *=  F_lang
        else:
            costarr[:] = [(u.installation_cost, u.utility_cost) for u in self._units]
            DPI_, UC_ = costarr.sum(0) 
            DC_ = DPI_ * (1 + o['Contingency'])
        FC_ = DC_ * (1 + o['Startup'] + o['Land'] + o['Royalties']) + o['Other fixed capital']
        MC_ = 0 # Material cost USD/hr
        S_  = 0 # Sales USD/hr
        for s in feeds:
            price = s.price
            if price: MC_ += price*s.massnet
        for s in products:
            price = s.price
            if price: S_ += price*s.massnet
        # Multiply by flow_factor for USD/yr
        UC_ *= flow_factor
        MC_ *= flow_factor
        S_ *= flow_factor
        FC_ = DC_*(1 + o['Startup'] + o['Land'] + o['Royalties']) + o['Other fixed capital']
        fb = o['Fringe benefits'] + o['Supplies']
        f =  (o['Maintenance']
            + o['Administration']
            + o['Property tax']
            + o['Property insurance'])
        WC_ = o['Working capital']*FC_
        L_ = o['Wage']*o['Employees']
        C_ = UC_ + (1+fb)*L_ + MC_ + f*FC_ + o['Other recurring costs']
        return DC_, FC_, WC_, S_, C_, o['Income tax'], UC_, MC_, L_
    
    @staticmethod
    def _calc_cashflow(cashflow_data,
                       parameters,
                       depreciation_data):
        """Perform cash flow analysis and return net present value."""
        # Cash flow data and parameters
        # C_DC: Depreciable capital
        # C_FC: Fixed capital
        # C_WC: Working capital
        # D: Depreciation
        # C: Annual operating cost (excluding depreciation)
        # S: Sales
        # NE: Net earnings
        # CF: Cash flow
        # DCF: Discounted cash flow
        # CPV: Cumulative present value
        C_DC, D, C_FC, C_WC, C, S, NE, CF, DCF, CPV = cashflow_data
        DC_, FC_, WC_, S_, C_, tax = parameters
        start, Depreciation, end, schedule = depreciation_data
        
        # Calculate
        D[:] = 0.0
        C_DC[:start] = DC_*schedule
        C_FC[:start] = FC_*schedule
        D[start:end] = DC_*Depreciation
        C_WC[start] = WC_
        C_WC[-1] = -WC_
        C[start:] = C_
        S[start:] = S_
        NE[:] = (S - C - D)*(1 - tax)
        CF[:] = (NE + D) - C_FC - C_WC
    
    @staticmethod
    def _calc_NPV(IRR, CF, duration_array):
        """Return NPV at given IRR and cashflow data."""
        return (CF/(1+IRR)**duration_array).sum()

    def solve_IRR(self):
        """Return the IRR at the break even point (NPV = 0) through cash flow analysis."""
        # Calculate cashflow table
        flow_factor, cashflow_info, depreciation_data = self._get_cached_data()
        cashflow_data, duration_array = cashflow_info
        parameters = self._calc_parameters(flow_factor)
        self._calc_cashflow(cashflow_data,
                            parameters[:-3],
                            depreciation_data)
        # Solve
        IRR = newton(self._calc_NPV, self._IRR_guess,
                     args=(cashflow_data[-3], duration_array),
                     tol=1e-5)
        self._IRR_guess = IRR if (0 < IRR < 1) else 0.15
        return IRR
    
    def solve_price(self, stream):
        """Return the price (USD/kg) of stream at the break even point (NPV = 0) through cash flow analysis. 
        
        **Parameters**
        
            **stream:** [Stream] Stream with variable selling price.
            
        """
        # Calculate cashflow table
        flow_factor, cashflow_info, depreciation_data = self._get_cached_data()
        cashflow_data, duration_array = cashflow_info
        parameters = self._calc_parameters(flow_factor)
        self._calc_cashflow(cashflow_data,
                            parameters[:-3],
                            depreciation_data)
        
        # Create function that adjusts cash flow with new stream price
        start = depreciation_data[0]
        tax = parameters[-4]
        IRR = self.options['IRR']
        cost_factor = stream.massnet*flow_factor*(1-tax)
        CF = cashflow_data[-3][start:]
        CF_copy = _array(CF)
        system = self.system
        if stream in system.feeds:
            adjust = CF_copy.__sub__
        elif stream in system.products:
            adjust = CF_copy.__add__
        else:
            raise ValueError(f"stream must be either a feed or a product of the system")
        
        # Solve
        calc_NPV = self._calc_NPV
        data_subset = cashflow_data[-3]
        def break_even_point(cost):
            CF[:] = adjust(cost)
            return calc_NPV(IRR, data_subset, duration_array)
        self._cost_guess = cost = newton(break_even_point, self._cost_guess)
        return stream.price + cost/cost_factor
    
    def __repr__(self):
        return f'<{type(self).__name__}: {self.system.ID}>'
    
    def _info(self):
        r = self._results()
        out = f'{type(self).__name__}: {self.system.ID}\n'
        IRR = self.options['IRR']*100
        if r:
            NPV = r['Net present value']
            ROI = r['Return on investment']
            PBP = r['Pay back period']
            out += f' NPV: {NPV:.3g} USD at {IRR:.1f}% IRR\n'
            out += f' ROI: {ROI:.3g} 1/yr\n'
            out += f' PBP: {PBP:.3g} yr' 
        else:
            out += f' NPV: None\n'
            out += f' ROI: None\n'
            out += f' PBP: None'
        return out
    
    def show(self):
        """Prints information on unit."""
        print(self._info())
    _ipython_display_ = show
        
        
    


# -*- coding: utf-8 -*-
"""
Created on Wed Jun 26 23:02:37 2019

@author: yoelr
"""
from .. import Unit

class SeedTrain(Unit):
    
    # Number of stages in series
    N_stages = 5
    
    #: Number of parallel seed trains
    N_trains = 2
    
    #: Cycle time for each batch (hr)
    batch_time = 24
    
    #: Turnaround time for each batch (hr)
    turnarround_time = 12
    
    #: wt % media (e.g. corn steep liquor) in each stage 
    media_loading = 0.50
    
    #: Diammonium phosphate loading in g/L of fermentation broth
    DAP = 0.67 
    
    #: Final fermentor volume (gal)
    maxvol = 2e5
    
    def _run(self): pass

    def _design(self): pass

    def _cost(self): pass

    


# -*- coding: utf-8 -*-
"""
Unit metaclasses are used to adjust mass and energy balance of Unit operations, mainly by setting the "_run" method.
"""
__all__ = []

from ._splitter import *
from ._final import *
from ._static import *
from ._mixer import *

from . import _splitter
from . import _final
from . import _static
from . import _mixer

__all__.extend(_splitter.__all__)
__all__.extend(_final.__all__)
__all__.extend(_static.__all__)
__all__.extend(_mixer.__all__)


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 14:26:41 2018

@author: yoelr
"""
from .. import Unit
from .metaclasses import final
from .metaclasses._mixer import run_mixer

class Mixer(Unit, metaclass=final):
    """Create a mixer that mixes any number of streams together.
    
    **Parameters**
    
    **ins**
    
        [:] Input streams
        
    **outs**
    
        [0] Mixed stream
    
    **Examples**
    
        Create a Mixer object with an ID and any number of input streams:
            
        .. code-block:: python
        
           >>> from biosteam import Species, Stream, Mixer
           >>> Stream.species = Species('Water', 'Ethanol')
           >>> s1 = Stream('s1', Water=20, T=340)
           >>> s2 = Stream('s2', Ethanol=10, T=300)
           >>> s3 = Stream('s3', Water=3, Ethanol=4)
           >>> M1 = Mixer('M1', ins=(s1, s2, s3), outs='s4')
           >>> M1.simulate()
           >>> M1.show()
           Mixer: M1
           ins...
           [0] s1
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water  20
           [1] s2
               phase: 'l', T: 300 K, P: 101325 Pa
               flow (kmol/hr): Ethanol  10
           [2] s3
               phase: 'l', T: 298.15 K, P: 101325 Pa
               flow (kmol/hr): Water    3
                               Ethanol  4
           outs...
           [0] s4
               phase: 'l', T: 317.54 K, P: 101325 Pa
               flow (kmol/hr): Water    23
                               Ethanol  14
           
    """
    results = None
    _N_ins = 2
    _N_outs = 1
    _has_cost = False
    
    def __init__(self, ID='', outs=(), ins=None):
        self.ID = ID
        self._init_ins(ins)
        self._init_outs(outs)

    simulate = _run = run_mixer
    installation_cost = purchase_cost = utility_cost = property(lambda: 0)




# -*- coding: utf-8 -*-
"""
Created on Thu May  9 13:38:57 2019

@author: Guest Group
"""
from ._block import Block
from .. import Unit, Stream
import numpy as np

__all__ = ('State',)

# %% functions

def param_unit(param):
    element = param.element
    if isinstance(element, Unit): return element
    elif isinstance(element, Stream): return element._sink

def parameter(system, element, setter, kind, name, distribution, units):
    if kind is 'coupled':
        return Block(element, system).parameter(setter, name=name, distribution=distribution, units=units)
    elif kind is 'isolated':
        return Block(element, None).parameter(setter, name=name, distribution=distribution, units=units)
    elif kind is 'design':
        return Block(element, None).parameter(setter, element._summary, name, distribution=distribution, units=units)
    elif kind is 'cost':
        if hasattr(element, '_end'):
            def simulate():
                element._cost()
                element._end()
        else:
            simulate = element._cost
        return Block(element, None).parameter(setter, simulate, name, distribution=distribution, units=units)
    raise ValueError(f"kind must be either 'coupled', 'isolated', 'design', or 'cost' (not {kind}).")

def modelfunction_with_skipping(params):
    cached = None
    zip_= zip
    params = tuple(params)
    def model(sample): 
        nonlocal cached
        try:
            sim = None
            for p, x, same in zip_(params, sample, cached==sample):
                if same: continue
                p.setter(x)
                if sim: continue
                if p.system: sim = p.simulate 
                else: p.simulate()
            if sim: sim()
            cached = sample
        except Exception as Error:
            cached = None
            raise Error
    return model

def modelfunction(params):
    zip_= zip
    params = tuple(params)
    def model(sample): 
        sim = None
        for p, x in zip_(params, sample):
            p.setter(x)
            if sim: continue
            if p.system: sim = p.simulate 
            else: p.simulate()
        if sim: sim()
    return model


# %%
    
class State:
    """Create a State object that can update the `system` state given a sample of parameter states.
    
    **Parameters**
    
        **system:** [System] Reflects the model state.
        
        **skip:** [bool] If True, skip simulation for repeated states
    
    """
    __slots__ = ('_system', # [System] Reflects the model state.
                 '_params', # list[Parameter] All parameters.
                 '_model', # [function] Model function.
                 '_skip') # [bool] If True, skip simulation for repeated states
    
    def __init__(self, system, skip=True):
        self._system = system
        self._params = []
        self._model = None
        self._skip = skip
    
    def __len__(self):
        return len(self._params)
    
    def get_parameters(self):
        if not self._model: self._loadmodel()
        return tuple(self._params)
    
    def parameter(self, setter=None, element=None, kind='isolated',
                  name=None, distribution=None, units=None):
        """Define parameter and add to model.
        
        **Parameters**
            
            **setter:** [function] Should set parameter in the element.
            
            **element:** [Unit or Stream] Element in the system being altered.
            
            **kind:** {'isolated', 'design', 'cost'}
                * 'coupled': parameter is coupled to the system.
                * 'isolated': parameter does not affect the system but does affect the element (if any).
                * 'design': parameter only affects design and cost of the element.
                * 'cost': parameter only affects cost of the element.
                
            **name:** [str] Name of parameter. If None, default to argument name of setter.
            
            **distribution:** [chaospy.Dist] Parameter distribution.
            
            **units:** [str] Parameter units
            
        .. Note::
            
            If kind is 'coupled', account for downstream operations. Otherwise, only account for given element. If kind is 'design' or 'cost', element must be a Unit object.
        
        """
        if not setter: return lambda setter: self.parameter(setter, element, kind, name, distribution, units)
        param = parameter(self._system, element, setter, kind, name, distribution, units)
        self._params.append(param)
        self._erase()
        return param
    
    def _erase(self):
        """Erase cached data."""
        self._model = None
    
    def _loadmodel(self):
        length = len( self._system._unitnetwork)
        index =  self._system._unitnetwork.index
        self._params.sort(key=lambda x: index(param_unit(x))
                                        if x.system else length)
        if self._skip:
            self._model = modelfunction_with_skipping(self._params)
        else:
            self._model = modelfunction(self._params)
    
    def __call__(self, sample):
        """Update state given sample of parameters."""
        if not self._model: self._loadmodel()
        return self._model(np.asarray(sample))
    
    def _repr(self):
        return f'{type(self).__name__}: {self._system}'
    
    def __repr__(self):
        return '<' + self._repr() + '>'
    
    def _info(self):
        if not self._model: self._loadmodel()
        if not self._params: return f'{self._repr()}\n (No parameters)'
        lines = []
        lenghts_block = []
        lastblk = None
        for i in self._params:
            blk = i.element_name
            element = len(blk)*' ' if blk==lastblk else blk
            lines.append(f" {element}${i.name}\n")
            lastblk = blk
            lenghts_block.append(len(blk))
        maxlen_block = max(lenghts_block)
        out = f'{self._repr()}\n'
        maxlen_block = max(maxlen_block, 7)
        out += ' Element:' + (maxlen_block - 7)*' ' + ' Parameter:\n'
        for newline, len_ in zip(lines, lenghts_block):
            newline = newline.replace('$', ' '*(maxlen_block-len_) + '  ')
            out += newline
        return out.rstrip('\n ')
    
    def show(self):
        """Return information on metric parameters."""
        print(self._info())
    _ipython_display_ = show
    


# -*- coding: utf-8 -*-
"""
Created on Thu Jun  6 00:55:03 2019

Equipment from the Humbird 2011 Report.
    
Humbird, D., Davis, R., Tao, L., Kinchin, C., Hsu, D., Aden, A., Dudgeon, D. (2011). Process Design and Economics for Biochemical Conversion of Lignocellulosic Biomass to Ethanol: Dilute-Acid Pretreatment and Enzymatic Hydrolysis of Corn Stover (No. NREL/TP-5100-47764, 1013269). https://doi.org/10.2172/1013269

@author: yoelr
"""
from .factories import xl2mod
import os
import sys
path = os.path.dirname(os.path.realpath(__file__)) + '\\'
xl2mod(path + '_humbird2011.xlsx', sys.modules[__name__])
del sys, xl2mod, os, path

# hp2kW = 0.7457

# # %% Decorators 

# _massflow_units = {'Flow rate': 'kg/hr'}
# def _design_kg_hr(self):
#     self._results['Design']['Flow rate'] = self._ins[0].massnet

# def cost_kg_hr(name=None, *, cost, exp, S, kW=0, CE=CE[2009], N=1):
#     def decorator(cls):
#         cls._design = _design_kg_hr
#         cls._units = _massflow_units
#         return decorators.cost('Flow rate', cost=cost, exp=exp,
#                                CE=CE, S=S, kW=kW, N=N)(cls)
#     return decorator

# def _design_Gcal_hr(self):
#     duty = self._outs[0].H - self._ins[0].H
#     self._heat_utilities[0](duty, self.ins[0].T, self._kwargs['T'])
#     self._results['Design']['Duty'] = duty*2.39e-7  # Gcal/hr

# _heatflow_units = {'Duty': 'Gcal/hr'}
# def heat_utility(name=None, *, cost, S, exp=0.7, kW=0, N=1, CE=CE[2009], BM=2.2):
#     """Decorate class as a heat exchanger."""
#     def decorator(cls):
#         cls.BM = BM
#         cls._graphics = units.HXutility._graphics
#         cls._linkedstreams = True
#         cls._N_heat_utilities = 1
#         cls._units = _heatflow_units
#         cls._design = _design_Gcal_hr
#         return decorators.cost('Duty', name, cost=cost, exp=exp,
#                                CE=CE, S=S, kW=kW, N=N)(cls)
#     return decorator
    

# # %% Units

# @cost_kg_hr(cost=13329690, exp=0.6, S=94697, kW=511.321)
# class FeedStockHandling(Unit):
#     """All area 100 equipment:
#         * C101 Transfer Conveyor (2)
#         * C102 High Angle Transfer Converyor (2)  
#         * C103 Reversing Load-in Conveyor
#         * C104 Dome Reclaim System (2)
#         * C106 High Angle Transfer Conveyor
#         * C107 Elevated Transfer Conveyor
#         * M101 Truck Scale (2)
#         * M102 Truck Dumper (2)
#         * M103 Truck Dumper Hopper (2)
#         * M104 Concrete Feedstock Storage Dome (2)
#         * M105 Belt Scale (2)
#         * M106 Dust Collection System (6)
#     """
#     _linkedstreams = True
#     BM = 1.7
    
# @cost_kg_hr(cost=6000, exp=0.5, S=136260)
# class SulfuricAcidMixer(Unit):
#     """A-201"""
#     _linkedstreams = True
#     BM = 1.0
    
# @cost_kg_hr(cost=19812400, exp=0.6, S=83333, kW=5290)
# class PretreatmentReactorSystem(Unit):
#     """Includes the following:
#         * C201 Transfer Conveyor (2)
#         * C202 Distribution Conveyor (2)
#         * C203 Overfeed Conveyor (4)
#         * C204 Pressurized Heating Screw
#         * C205 Pressurized Pre-heater Discharge (2)
#         * C206 Pressurized Transport #1
#         * C207 Pressurized Transport #2
#         * M201 Doffing Roll Storage Bings (2)
#         * M202 Pin Drum Feeder (2)
#         * M203 Plug Screw Feeder (2)
#         * M204 Prehydrolysis / Vertical Preheater
#         * M205 Pin Drum Feeder (2)
#         * M206 Plug Screw Feeder  (2)
#         * M207 Pretreatment Reactor (3)
#     """
#     _linkedstreams = True
#     BM = 1.5
#     def _init(self):
#         self._water_mass = Stream.indices('Water')
    
#     def _design(self):
#         feed = self._ins[0]
#         self._results['Design']['Flow rate'] = feed.massnet - feed.mass[self._water_index]
        

# @heat_utility(cost=92e3, CE=CE[2010], S=-8, exp=0.70)
# class PretreatmentWaterHeater(Unit):
#     """H201"""
#     _kwargs = {'T': None}
    
#     def _run(self):
#         out = self._outs[0]
#         out.P = self._ins[0].P
#         out.T = self._kwargs['T']
    
    
# @heat_utility(cost=34e3, S=2, exp=0.70)
# class WasteVaporCondenser(Unit):
#     """H244"""
#     _kwargs = {'T': None}
#     def _setup(self):
#         self._outs[0].phase = 'l'
    
#     def _run(self):
#         feed = self._ins[0]
#         out = self._outs[0]
#         out.P = feed.P
#         out.T = self._kwargs['T'] or feed.T

# @decorators.cost('Flow rate', 'Discharge pump', kW=75*hp2kW, CE=CE[2009], cost=30e3, exp=0.80)
# @cost_kg_hr('Agitators', N=3, cost=90e3/3, S=252891, exp=0.5)
# class Flash204(units.Flash):
#     """Includes:
#         * Discharge pump
#         * Agitator
#     """
#     BM = {'Discharge pump': 1.5,
#           'Agitators': 2.3}
    
    
    
        


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:17:05 2018

@author: yoelr
"""
from .metaclasses import splitter
from .. import Unit
from .decorators import cost

__all__ = ('MolecularSieve',)


@cost('Flow rate', 'Pressure filter drying (2)',
      cost=405000, CE=521.9, S=22687, n=0.6, kW=1044)
@cost('Flow rate', 'Pressure filter pressing',
      cost=75200, CE=521.9, S=22687, n=0.6, kW=112)
@cost('Flow rate', 'Column', units='kg/hr',
      cost=2601000, CE=521.9, S=22687, n=0.6)
class MolecularSieve(Unit, metaclass=splitter):
    """Create an ethanol/water molecular sieve for bioethanol plants.
    The molecular sieve is modeled as a component wise separator. Costing
    is based on scaling by the 6/10ths rule from an NREL TEA report [1].
    
    **Parameters**
    
        **split:** [array_like] Componentwise split to the 0th output stream

    **ins**
    
        [0] Feed (gas)
        
    **outs**
    
        [0] Split stream (gas)
        
        [1] Remainder stream (gas)
    
    **References**
    
        [1] Process Design and Economics for Biochemical Conversion of
        Lignocellulosic Biomass to Ethanol Dilute-Acid Pretreatment and
        Enzymatic Hydrolysis of Corn Stover. D. Humbird, R. Davis, L.
        Tao, C. Kinchin, D. Hsu, and A. Aden (National Renewable Energy
        Laboratory Golden, Colorado). P. Schoen, J. Lukas, B. Olthof,
        M. Worley, D. Sexton, and D. Dudgeon (Harris Group Inc. Seattle,
        Washington and Atlanta, Georgia)
    
    **Examples**
    
        :doc:`MolecularSieve Example`
    
    """
    BM = 1.8
    def _init(self):
        self._outs[0]._phase = self._outs[1]._phase = 'g'





# -*- coding: utf-8 -*-
"""
Created on Sat Nov 18 16:17:00 2017

@author: Yoel Cortes-Pena
"""
from setuptools import setup
#from Cython.Build import cythonize
#import numpy

setup(
    name='biosteam',
    packages=['biosteam'],
    license='MIT',
    version='0.5.0',
    description='The Biorefinery Simulation and Techno-Economic Analysis Modules',
    long_description=open('README.rst').read(),
    #ext_modules=cythonize('biosteam/equilibrium/unifac.pyx'),
    #include_dirs=[numpy.get_include()],
    author='Yoel Cortes-Pena',
    install_requires=['pint==0.9', 'ht==0.1.52', 'fluids==0.1.74',
                      'scipy', 'IPython', 'thermo==0.1.39', 'colorpalette==0.1.1',
                      'array_collections==0.1.9', 'free_properties==0.1.9',
                      'pandas', 'numpy', 'graphviz', 'matplotlib'],
    python_requires=">=3.6",
    package_data=
        {'biosteam': ['_equilibrium/*', 'inspect/*', 'price/*', '_report/*',
                      'my_units_defs.txt', '_utils/*', 'units/*',
                      '_units/_humbird2011.xlsx', 'evaluation/*', 
                      'compounds/*', 'units/designtools/*', 'units/metaclasses/*',
                      'units/facilities/*', 'units/decorators/*', 'reaction/*'], },
    platforms=["Windows", "Mac"],
    author_email='yoelcortes@gmail.com',
    url='https://github.com/yoelcortes/biosteam',
    download_url='https://github.com/yoelcortes/biosteam.git',
    classifiers=['Development Status :: 3 - Alpha',
                 'Environment :: Console',
                 'License :: OSI Approved :: MIT License',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Topic :: Scientific/Engineering',
                 'Topic :: Scientific/Engineering :: Chemistry',
                 'Topic :: Scientific/Engineering :: Mathematics'],
    keywords='chemical process simmulation bioprocess engineering mass energy balance material properties phase equilibrium CABBI biorefinery biofuel bioproducts',
)


# -*- coding: utf-8 -*-
"""
Created on Sat Aug 18 15:04:55 2018

@author: yoelr
"""
from copy import copy
import numpy as np
from scipy.optimize import newton
from ._exceptions import SolverError, _try_method
from ._flowsheet import find, make_digraph, save_digraph
from ._stream import Stream
from ._unit import Unit
from ._report import save_report
from ._utils import color_scheme, missing_stream, strtuple, function
import biosteam as bst

__all__ = ('System',)

# %% Debugging and exception handling

def _evaluate(self, command=None):
    """Evaluate a command and request user input for next command. If no command, return. This function is used for debugging a System object."""    
    CS = color_scheme
    # Done evaluating if no command, exit debugger if 'exit'
    if command is None:
        Next = CS.next('Next: ') + f'{repr(self)}\n'
        info = CS.info("Enter to continue or type to evaluate:\n")
        command = input(Next + info + ">>> ")
    
    if command == 'exit': raise KeyboardInterrupt()
    if command:
        # Build locals dictionary for evaluating command
        lcs = {} 
        for attr in ('stream', 'unit', 'system'):
            dct = getattr(find, attr).__dict__
            lcs.update({i:j() for i, j in dct.items()})
        lcs.update({attr:getattr(bst, attr) for attr in bst.__all__})
        try:
            out = eval(command, {}, lcs)            
        except Exception as err:
            # Print exception and ask to raise error or continue evaluating
            err = CS.exception(f'{type(err).__name__}:') + f' {str(err)}\n\n'
            info = CS.info(f"Enter to raise error or type to evaluate:\n")
            command = input(err + info + ">>> ")
            if command == '': raise err
            _evaluate(self, command)        
        else:
            # If successful, continue evaluating
            if out is None: pass
            elif (not hasattr(out, '_ipython_display_')
                  or isinstance(out, type)): print(out)
            else: out._ipython_display_()
                
            command = input(">>> ")
            _evaluate(self, command)

def _method_debug(self, func):
    """Method decorator for debugging system."""
    def wrapper(*args, **kwargs):
        # Run method and ask to evaluate
        _evaluate(self)
        func(*args, **kwargs)
        
    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    wrapper._original = func
    return wrapper

def _notify_run_wrapper(self, func):
    """Decorate a System run method to notify you after each loop"""
    def wrapper(*args, **kwargs):
        if self.recycle:
            func(*args, **kwargs)
            x = self._solver_error['iter']
            input(f'        Finished loop #{x}\n')
        else:
            func(*args, **kwargs)
    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    wrapper._original = func
    return wrapper
    
# %% System node for diagram

class _systemUnit(Unit):
    """Dummy unit for displaying a system."""
    line = 'System'
    ID = None
    _ID = property(lambda self: self.ID)

_sysgraphics = _systemUnit._graphics
_sysgraphics.edge_in = _sysgraphics.edge_in * 10
_sysgraphics.edge_out = _sysgraphics.edge_out * 15
_sysgraphics.node['peripheries'] = '2'

class _streamUnit(Unit):
    line = ''
    ID = None
    _ID = _systemUnit._ID
    
_stream_graphics = _streamUnit._graphics
_stream_graphics.node['fillcolor'] = 'white:#79dae8'
del _stream_graphics, _sysgraphics

# %% Other

_isfeed = lambda stream: not stream._source and stream._sink
_isproduct = lambda stream: not stream._sink and stream._source


# %% Process flow

class System:
    """Create a System object that can iteratively run each element in a network of BioSTREAM objects until the recycle stream is converged. A network can have function, Unit and/or System objects. When the network contains an inner System object, it converges/solves it in each loop/iteration.

    **Parameters**

         **ID:** [str] A unique identification. If ID is None, instance will not be registered in flowsheet.

         **network:** tuple[Unit, function and/or System] A network that is run element by element until the recycle converges.

         **recycle:** [Stream] A tear stream for the recycle loop.
         
         **facilities:** tuple[Unit, function, and/or System] Offsite facilities that are simulated only after completing the network simulation.

    """
    ### Class attributes ###
    
    #: [dict] Dictionary of convergence options regarding maximum number of iterations and molar flow rate and temperature tolerances
    options = {'Maximum iteration': 100,
               'Molar tolerance (kmol/hr)': 0.01,
               'Temperature tolerance (K)': 0.10}

    # [dict] Cached downstream systems by (system, unit, with_facilities) keys
    _cached_downstream_systems = {} 

    def __init__(self, ID, network, recycle=None, facilities=()):
        #: [dict] Current molar flow and temperature errors and number of iterations made
        self._solver_error = {'mol_error': 0,
                              'T_error': 0,
                              'spec_error': 0,
                              'iter': 0}
         
        #: set[Stream] All streams within the system
        self.streams = streams = set()

        #: set[System] All subsystems in the system
        self.subsystems = subsystems = set()
        
        #: list[Unit] Network of only unit operations
        self._unitnetwork = units = []
        
        #: tuple[Unit, function and/or System] A network that is run element by element until the recycle converges.
        self.network = tuple(network)
        inst = isinstance
        for i in network:
            if i in units: continue
            if inst(i, Unit): 
                units.append(i)
                streams.update(i._ins)
                streams.update(i._outs)
            elif inst(i, System):
                units.extend(i._unitnetwork)
                subsystems.add(i)
                streams.update(i.streams)
        streams.discard(missing_stream) 
        
        # link all unit operations with linked streams
        has = hasattr
        for u in units:
            if has(u, '_link_streams'): u._link_streams()
        
        #: set[Unit] All units within the system
        self.units = units = set(units)
        
        #: set[Unit] All units in the network that have costs
        self._network_costunits = costunits = {i for i in units if i._has_cost}
        
        #: set[Unit] All units that have costs.
        self._costunits = costunits = costunits.copy()
        
        #: tuple[Unit, function, and/or System] Offsite facilities that are simulated only after completing the network simulation.
        self.facilities = tuple(facilities)
        for i in facilities:
            if inst(i, Unit):
                units.add(i)
                streams.update(i._ins + i._outs)
                if i._has_cost: costunits.add(i)
            elif inst(i, System):
                units.update(i.units)
                streams.update(i.streams)
                subsystems.add(i)
                costunits.update(i._costunits)
        
        #: set[Stream] All feed streams in the system.
        self.feeds = set(filter(_isfeed, streams))
        
        #: set[Stream] All product streams in the system.
        self.products = set(filter(_isproduct, streams)) 
        
        #: [TEA] System object for Techno-Economic Analysis.
        self._TEA = None
        self.recycle = recycle
        
        if ID:
            ID = ID.replace(' ', '_')
            ID_words = ID.split('_')
            if not all(word.isalnum() for word in ID_words):
                raise ValueError('ID cannot have any special characters')
            self._ID = ID
            find.system[ID] = self
    
    save_report = save_report
    
    @property
    def ID(self):
        """Identification."""
        return self._ID
    
    @property
    def TEA(self):
        """[TEA] System object for Techno-Economic Analysis."""
        return self._TEA
    
    @property
    def recycle(self):
        """A tear stream for the recycle loop"""
        return self._recycle

    @recycle.setter
    def recycle(self, stream):
        if stream is None:
            self._converge_method = self._run
        elif not isinstance(stream, Stream):
            raise ValueError(f"recycle must be a Stream instance or None, not {type(stream).__name__}")
        self._recycle = stream

    @property
    def converge_method(self):
        """Iterative convergence method ('Wegstein', or 'fixed point')."""
        return self._converge_method.__name__[1:]

    @converge_method.setter
    def converge_method(self, method):
        if self.recycle is None:
            raise ValueError("cannot set converge method when no recyle is specified")
        method = method.lower().replace('-', '').replace(' ', '')
        if 'wegstein' in method:
            self._converge_method = self._Wegstein
        elif 'fixedpoint' in method:
            self._converge_method = self._fixed_point
        else:
            raise ValueError(f"only 'Wegstein' and 'fixed point' methods are valid, not '{method}'")

    
    def _downstream_network(self, unit):
        """Return a list composed of the `unit` and everything downstream."""
        if unit not in self.units: return []
        elif self._recycle: return self.network
        unit_found = False
        downstream_units = unit._downstream_units
        network = []
        inst = isinstance
        for i in self.network:
            if unit_found:
                if inst(i, System):
                    for u in i.units:
                        if u in downstream_units:
                            network.append(i)
                            break
                elif i in downstream_units:
                    network.append(i)
                elif (not inst(i, Unit)
                      or i.line == 'Balance'):
                    network.append(i)
            else:
                if unit is i:
                    unit_found = True
                    network.append(unit)
                elif inst(i, System) and unit in i.units:
                        unit_found = True   
                        network.append(i)
        return network

    def _downstream_system(self, unit):
        """Return a system with a network composed of the `unit` and everything downstream (facilities included)."""
        if unit is self.network[0]: return self
        system = self._cached_downstream_systems.get((self, unit))
        if system: return system
        network = self._downstream_network(unit)
        if network:
            downstream_facilities = self.facilities            
        else:
            unit_not_found = True
            inst = isinstance
            for pos, i in enumerate(self.facilities):
                if unit is i or (inst(i, System) and unit in i.units):
                    downstream_facilities = self.facilities[pos:]
                    unit_not_found = False
                    break
            if unit_not_found: raise ValueError(f'{unit} not found in system')
        system = System(None, network,
                        facilities=downstream_facilities)
        system._ID = f'{type(unit).__name__}-{unit} and downstream'
        self._cached_downstream_systems[unit] = system
        return system
    
    def _minimal_diagram(self, file, format='svg'):
        """Minimally display the network as a box."""
        outs = []
        ins = []
        for s in self.streams:
            source = s._source
            sink = s._sink
            if source in self.units and sink not in self.units:
                outs.append(s)
            elif sink in self.units and source not in self.units:
                ins.append(s)
        product = Stream(None)
        product._ID = ''
        feed = Stream(None)
        feed._ID = ''
        _streamUnit('\n'.join([i.ID for i in ins]),
                    feed)
        _streamUnit('\n'.join([i.ID for i in outs]),
                    None, product)
        unit = _systemUnit(self.ID, product, feed)
        unit.diagram(1, file, format)

    def _surface_diagram(self, file, format='svg'):
        """Display only surface elements listed in the network."""
        # Get surface items to make nodes and edges
        units = set()  
        refresh_units = set()
        for i in self.network:
            if isinstance(i, Unit):
                units.add(i)
            elif isinstance(i, System):
                outs = []
                ins = []
                feeds = []
                products = []
                for s in i.streams:
                    source = s._source
                    sink = s._sink
                    if source in i.units and sink not in i.units:
                        if sink: outs.append(s)
                        else: products.append(s)
                        refresh_units.add(source)
                    elif sink in i.units and source not in i.units:
                        if source: ins.append(s)
                        else: feeds.append(s)
                        refresh_units.add(sink)
                
                if len(feeds) > 1:
                    feed = Stream(None)
                    feed._ID = ''
                    units.add(_streamUnit('\n'.join([i.ID for i in feeds]), feed))
                    ins.append(feed)
                else: ins += feeds
                
                if len(products) > 1:
                    product = Stream(None)
                    product._ID = ''
                    units.add(_streamUnit('\n'.join([i.ID for i in products]),
                                          None, product))
                    outs.append(product)
                else: outs += products
                
                subsystem_unit = _systemUnit(i.ID, outs, ins)
                units.add(subsystem_unit)
                
        System(None, units)._thorough_diagram(file, format)
        # Reconnect how it was
        for u in refresh_units:
            u._ins[:] = u._ins
            u._outs[:] = u._outs
      
    def _thorough_diagram(self, file=None, format='svg'):
        """Thoroughly display every unit within the network."""
        # Create a digraph and set direction left to right
        f = make_digraph(self.units, self.streams)
        save_digraph(f, file, format)
        
    def diagram(self, kind='surface', file=None, format='svg'):
        """Display a `Graphviz <https://pypi.org/project/graphviz/>`__ diagram of the system.
        
        **Parameters**
        
            **kind:** Must be one of the following:
                * **'thorough':** Thoroughly display every unit within the network
                * **'surface':** Display only surface elements listed in the network
                * **'minimal':** Minimally display the network as a box
        
            **file:** Must be one of the following:
                * [str] File name to save diagram.
                * [None] Display diagram in console.
        
            **format:** File format.
        
        """
        if kind == 'thorough':
            return self._thorough_diagram(file, format)
        elif kind == 'surface':
            return self._surface_diagram(file, format)
        elif kind == 'minimal':
            return self._minimal_diagram(file, format)
        else:
            raise ValueError(f"kind must be either 'thorough', 'surface', or 'minimal'")
            

    # Methods for running one iteration of a loop
    def _run(self):
        """Rigorous run each element of the system."""
        inst = isinstance
        _try = _try_method
        for a in self.network:
            if inst(a, Unit): _try(a._run)
            elif inst(a, System): a._converge()
            else: a() # Assume it is a function
        self._solver_error['iter'] += 1
    
    # Methods for convering the recycle stream
    def _fixed_point(self):
        """Converge system recycle using inner and outer loops with fixed-point iteration."""
        # Reused attributes
        recycle = self.recycle
        run = self._run
        solver_error = self._solver_error
        maxiter, mol_tol, T_tol = self.options.values()

        # Outer loop
        abs_ = abs
        while True:
            mol_old = copy(recycle.mol)
            T_old = recycle.T
            run()
            mol_error = abs_(recycle.mol - mol_old).sum()
            T_error = abs_(recycle.T - T_old)
            if T_error < T_tol and mol_error < mol_tol: break
            if solver_error['iter'] > maxiter:
                solver_error['mol_error'] = mol_error
                solver_error['T_error'] = T_error
                raise SolverError(f'could not converge'
                                  + self._error_info())

        solver_error['mol_error'] = mol_error
        solver_error['T_error'] = T_error

    def _Wegstein(self):
        """Converge the system recycle iteratively using Wegstein's method."""
        # Reused attributes
        recycle = self.recycle
        run = self._run
        maxiter, mol_tol, T_tol = self.options.values()
        solver_error = self._solver_error

        # Prepare variables
        len_ = recycle._species._Nspecies + 1
        x0 = np.zeros(len_)
        gx0 = np.zeros(len_)
        x1 = np.zeros(len_)
        gx1 = np.zeros(len_)
        ones = np.ones(len_)
        s = np.zeros(len_)

        # First run
        x0[:-1] = recycle.mol
        x0[-1] = recycle.T
        run()
        x1[:-1] = gx0[:-1] = recycle.mol
        x1[-1] = gx0[-1] = recycle.T

        # Check convergence
        abs_ = abs
        mol_error = abs_(gx0[:-1] - x0[:-1]).sum()
        T_error = abs_(gx0[-1] - x0[-1])
        converged = mol_error < mol_tol and T_error < T_tol

        # Outer loop
        while not converged:
            run()
            gx1[:-1] = recycle.mol
            gx1[-1] = recycle.T

            # Check if converged
            mol_error = abs_(gx1[:-1] - x1[:-1]).sum()
            T_error = abs_(gx1[-1] - x1[-1])
            converged = mol_error < mol_tol and T_error < T_tol

            if solver_error['iter'] > maxiter:
                solver_error['mol_error'] = mol_error
                solver_error['T_error'] = T_error
                raise SolverError(f'{self} could not converge'
                                  + self._error_info())

            # Get relaxation factor and set next iteration
            x_diff = x1 - x0
            pos = x_diff != 0
            s[pos] = (gx1[pos] - gx0[pos])/x_diff[pos]
            x0 = x1
            gx0 = gx1
            s[s > 0.9] = 0.9
            w = ones/(ones-s)
            x1 = w*gx1 + (1-w)*x1
            recycle._mol[:] = x1[:-1]
            recycle.T = x1[-1]
        solver_error['mol_error'] = mol_error
        solver_error['T_error'] = T_error
    
    # Default converge method
    _converge_method = _Wegstein
    
    def _converge(self):
        """Converge the system recycle using an iterative solver (Wegstein by default)."""
        try:
            return self._converge_method()
        except Exception as e:
            if self._recycle: self._recycle.empty()
            self._converge_method()

    def set_spec(self, getter, setter, solver=newton, **kwargs):
        """Wrap a solver around the converge method.

        **Parameters**

             **getter:** [function] Returns objective value.

             **setter:** [function] Changes independent variable.
             
             **solver:** [function] Solves objective function.

             **kwargs:** [dict] Key word arguments passed to solver.

        """
        converge = self._converge_method
        solver_error = self._solver_error

        def error(val):
            setter(val)
            converge()
            solver_error['spec_error'] = e = getter()
            return e

        def converge_spec():
            x = solver(error, **kwargs)
            if solver is newton: kwargs['x0'] = x

        self._converge_method = converge_spec

    def _reset_iter(self):
        self._solver_error['iter'] = 0
        for system in self.subsystems: system._reset_iter()
    
    def reset_names(self, unit_format=None, stream_format=None):
        """Reset names of all streams and units according to the network order."""
        Unit._default_ID = unit_format if unit_format else ['U', 0]
        Stream._default_ID = stream_format if stream_format else ['d', 0]
        streams = set()
        units = set()
        for i in self._unitnetwork:
            if i in units: continue
            try: i.ID = ''
            except: continue
            for s in (i._ins + i._outs):
                if (s is not missing_stream
                    and s._sink and s._source
                    and s not in streams):
                    s.ID = ''
                    streams.add(s)
            units.add(i)
    
    def reset_flows(self):
        """Reset all process streams to zero flow."""
        self._solver_error = {'mol_error': 0,
                              'T_error': 0,
                              'spec_error': 0,
                              'iter': 0}
        feeds = self.feeds
        for stream in self.streams:
            if stream._link: continue
            if stream not in feeds: stream.empty()

    def simulate(self):
        """Converge the network and simulate all units."""
        self._reset_iter()
        self._converge()
        _try = _try_method
        for u in self._network_costunits: _try(u._summary)
        inst = isinstance
        for i in self.facilities:
            if inst(i, (Unit, System)): i.simulate()
            else: i() # Assume it is a function
        
    # Debugging
    def _debug_on(self):
        """Turn on debug mode."""
        self._run = _notify_run_wrapper(self, self._run)
        self.network = network = list(self.network)
        for i, item in enumerate(network):
            if isinstance(item, Unit):
                item._run = _method_debug(item, item._run)
            elif isinstance(item, System):
                item._converge = _method_debug(item, item._converge)
            elif isinstance(item, function):
                network[i] = _method_debug(item, item)

    def _debug_off(self):
        """Turn off debug mode."""
        self._run = self._run._original
        network = self.network
        for i, item in enumerate(network):
            if isinstance(item, Unit):
                item._run = item._run._original
            elif isinstance(item, System):
                item._converge = item._converge._original
            elif isinstance(item, function):
                network[i] = item._original
        self.network = tuple(network)
    
    def debug(self):
        """Convege in debug mode. Just try it!"""
        self._debug_on()
        try:
            self._converge_method()
        except Exception as error:
            self._debug_off()
            raise error
        else:
            end = self._error_info()
            if end:
                print(f'\nFinished debugging{end}')
            else:
                print(f'\n        Finished debugging')
            self._debug_off()

    # Representation
    def __str__(self):
        if self.ID: return self.ID
        else: return type(self).__name__ 
    
    def __repr__(self):
        if self.ID: return f'<{type(self).__name__}: {self.ID}>'
        else: return f'<{type(self).__name__}>'

    def show(self):
        """Prints information on unit."""
        print(self._info())
    
    def _ipython_display_(self):
        try: self.diagram('minimal')
        except: pass
        self.show()

    def _error_info(self):
        """Return information on convergence."""
        x = self._solver_error
        recycle = self.recycle
        error_info = ''
        spec = x['spec_error']
        if spec:
            error_info += f'\n specification error: {spec:.3g}'
        if recycle:
            error_info += f"\n convergence error: Flow rate   {x['mol_error']:.2e} kmol/hr"
            error_info += f"\n                    Temperature {x['T_error']:.2e} K"
        if spec or recycle:
            error_info += f"\n iterations: {x['iter']}"
        return error_info

    def _info(self):
        """Return string with all specifications."""
        if self.recycle is None:
            recycle = ''
        else:
            recycle = f"\n recycle: {self.recycle}"
        error = self._error_info()
        network = strtuple(self.network)
        i = 1; last_i = 0
        while True:
            i += 2
            i = network.find(', ', i)
            i_next = network.find(', ', i+2)
            if (i_next-last_i) > 35:
                network = (network[:i] + '%' + network[i:])
                last_i = i
            elif i == -1: break
        network = network.replace('%, ', ',\n'+' '*11)
        
        if self.facilities:
            facilities = strtuple(self.facilities)
            i = 1; last_i = 0
            while True:
                i += 2
                i = facilities.find(', ', i)
                if (i - last_i) > 35:
                    facilities = (facilities[:i] + '%' + facilities[i:])
                    last_i = i
                elif i == -1: break
            facilities = facilities.replace('%, ', ',\n'+' '*14)
            facilities = f"\n facilities: {facilities}" 
        else:
            facilities = ''
        
        return (f"System: {self.ID}"
                + recycle
                + f"\n network: {network}"
                + facilities
                + error)



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 14:34:07 2018

@author: yoelr
"""
from .. import Unit, Stream
from .metaclasses import final
from numpy import asarray
from pandas import Series
from .metaclasses._splitter import split
from . import Mixer

class Splitter(Unit, metaclass=final):
    """Create a splitter that separates mixed streams based on splits.

    **Parameters**

        **split:** Should be one of the following
            * [float] The fraction of net feed in the 0th output stream
            * [array_like] Componentwise split of feed to 0th output stream
            
        **order:** Iterable[str] Species order of split
    
    **ins**
    
        [0] Feed stream
        
    **outs**
    
        [0] Split stream
        
        [1] Remainder stream
    
    **Examples**
    
        Create a Splitter object with an ID, a feed stream, two output streams, and an overall split:
            
        .. code-block:: python
        
           >>> from biosteam import Species, Stream, units
           >>> Stream.species = Species('Water', 'Ethanol')
           >>> feed = Stream('feed', Water=20, Ethanol=10, T=340)
           >>> S1 = units.Splitter('S1', ins=feed, outs=('top', 'bot'), split=0.1)
           >>> S1.simulate()
           >>> S1.show()
           Splitter: S1
           ins...
           [0] feed
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    20
                               Ethanol  10
           outs...
           [0] top
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    2
                               Ethanol  1
           [1] bot
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    18
                               Ethanol  9
          
        Create a Splitter object, but this time with a componentwise split:
            
        .. code-block:: python
        
           >>> from biosteam import Species, Stream, units
           >>> Stream.species = Species('Water', 'Ethanol')
           >>> feed = Stream('feed', Water=20, Ethanol=10, T=340)
           >>> S1 = units.Splitter('S1', ins=feed, outs=('top', 'bot'),
           ...                     split=(0.1, 0.99))
           >>> S1.simulate()
           >>> S1.show()
           Splitter: S1
           ins...
           [0] feed
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    20
                               Ethanol  10
           outs...
           [0] top
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    2
                               Ethanol  9.9
           [1] bot
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    18
                               Ethanol  0.1
                               
        Create a Splitter object using componentwise split, but this time specify the order:
            
        .. code-block:: python
        
           >>> from biosteam import Species, Stream, units
           >>> Stream.species = Species('Water', 'Ethanol')
           >>> feed = Stream('feed', Water=20, Ethanol=10, T=340)
           >>> S1 = units.Splitter('S1', ins=feed, outs=('top', 'bot'),
           ...                     split=(0.99, 0.01), order=('Ethanol', 'Water'))
           >>> S1.simulate()
           >>> S1.show()
           Splitter: S1
           ins...
           [0] feed
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    20
                               Ethanol  10
           outs...
           [0] top
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    2
                               Ethanol  9.9
           [1] bot
               phase: 'l', T: 340 K, P: 101325 Pa
               flow (kmol/hr): Water    18
                               Ethanol  0.1
    
    """
    results = None
    _has_cost = False
    _N_outs = 2

    def __init__(self, ID='', outs=(), ins=None, split=None, order=None):
        self.ID = ID
        self._split = split = Stream.species.array(order, split) if order else asarray(split)
        self._split_series =  Series(split, Stream.species.IDs)
        self._init_ins(ins)
        self._init_outs(outs)

    def _run(self):
        top, bot = self._outs
        feed = self._ins[0]
        net_mol = feed.mol
        bot.T = top.T = feed.T
        bot.P = top.P = feed.P
        bot._phase = top._phase = feed._phase
        top._mol[:] = net_mol * self._split
        bot._mol[:] = net_mol - top._mol
    
    split = split
    simulate = _run
    installation_cost = purchase_cost = utility_cost = Mixer.purchase_cost


class InvSplitter(Unit, metaclass=final):
    """Create a splitter that sets the input stream based on output streams. Must have only one input stream. The output streams will become the same temperature, pressure and phase as the input.
    """
    line = 'Splitter'
    results = None
    _graphics = Splitter._graphics
    _has_cost = False
    
    def __init__(self, ID='', outs=(), ins=None):
        self.ID = ID
        self._init_ins(ins)
        self._init_outs(outs)
        
    def _run(self):
        feed = self.ins[0]
        feed._mol[:] = self._mol_out
        T = feed.T
        P = feed.P
        phase = feed.phase
        for out in self._outs:
            out.T = T
            out.P = P
            out.phase = phase 
            
    simulate = _run





# -*- coding: utf-8 -*-
"""
Created on Sat Jun 15 21:01:26 2019

@author: yoelr
"""
from ..._unit import Outs, Stream, missing_stream, metaUnit, Unit, _try_method

__all__ = ('static',)

def _init_outs(self, outs):
    """Initialize output streams."""
    if outs is None:
        self._outs = Outs(self, (Stream.proxy(None) for i in self._N_outs))
    elif not outs:
        self._outs = Outs(self, (Stream.proxy('') for i in self._ins))
    elif isinstance(outs, Stream):
        self._outs = Outs(self, (outs,))
    elif isinstance(outs, str):
        self._outs = Outs(self, (Stream.proxy(outs),))
    else:
        self._outs = Outs(self, (o if isinstance(o, Stream)
                                 else Stream.proxy(o)
                                 for o in outs))
        
def _link_streams(self):
    """Link product to feed."""
    try: self._outs[0].link = self._ins[0]
    except Exception as Error:
        if missing_stream in (self._ins + self._outs):
            raise AttributeError(f'missing stream object in {repr(self)}')
   
def simulate(self):
    """Run rigourous simulation and determine all design requirements and costs."""
    self._link_streams()
    _try_method(self._run)
    _try_method(self._summary)

def _proxy_run(self):
    o = self._outs[0]
    i = self._ins[0]
    o.T = i.T
    o.P = i.P
    o._phase = i._phase
        
def _info(self, T, P, flow, fraction):
    """Information on unit."""
    self._link_streams()
    return Unit._info(self, T, P, flow, fraction)
    
class static(metaUnit):
    """Unit subclasses that implement the static metaclass expect only one input stream (the `feed`) and only one output stream (the `product`). The static metaclass makes the `feed` and `product` share data on material flow rates. Therefore `feed`.mol and `product`.mol will always be equal. Also, the `product`'s T, P, and phase will be equal to the `feed`'s upon simulation if no `_run` method is implemented."""
    def __new__(mcl, name, bases, dct):
        try:
            base, = bases
        except:
            TypeError('cannot create {mcl.__name__} subclass from more than one super class')
           
        cls = super().__new__(mcl, name, bases, dct)
        if not isinstance(base, mcl):
            cls._init_outs = _init_outs
            cls._N_ins = cls._N_outs = 1
            cls._link_streams = _link_streams
            cls.simulate = simulate
            if '_run' not in dct: cls._run = _proxy_run
        return cls
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 16 11:49:52 2018

@author: Yoel Rene Cortes-Pena
"""
from . import array_utils
from . import color_utils
from . import other_utils
from . import stream_utils
from . import display_units
from . import register

from .array_utils import *
from .color_utils import *
from .other_utils import *
from .stream_utils import *
from .display_units import *
from .register import *

__all__ = []
__all__.extend(display_units.__all__)
__all__.extend(array_utils.__all__)
__all__.extend(color_utils.__all__)
__all__.extend(other_utils.__all__)
__all__.extend(stream_utils.__all__)
__all__.extend(register.__all__)


# -*- coding: utf-8 -*-
"""
Created on Tue May 14 14:20:53 2019

@author: yoelr
"""
__all__ = ('Parameter',)
from ._name import elementname

class Parameter:
    """Create a Parameter object that, when called, runs the setter and the simulate functions.
    
    **Parameters**
    
        **name:** [str] Name of parameter
        
        **setter:** [function] Should set the parameter
        
        **simulate:** [function] Should simulate parameter effects
        
        **element:** [object] Element associated to parameter
        
        **system:** [System] System associated to parameter
        
        **distribution:** [chaospy Dist] Parameter distribution
        
        **units:** [str] Units of parameter
    
    """
    __slots__ = ('name', 'setter', 'simulate', 'element', 'system', 'distribution', 'units')
    
    def __init__(self, name, setter, simulate,
                 element, system, distribution,
                 units):
        self.name = name.replace('_', ' ').capitalize()
        self.setter = setter
        self.simulate = simulate
        self.element = element
        self.system = system
        self.distribution = distribution
        self.units = units
    
    def __call__(self, value):
        self.setter(value)
        self.simulate()
    
    @property
    def element_name(self):
        return elementname(self.element)
    
    def __repr__(self):
        units = f" ({self.units})" if self.units else ""
        element = f" [{self.element_name}]" if self.element else ""
        return f'<{type(self).__name__}:{element} {self.name}{units}>'
    
    def show(self):
        print(f'{type(self).__name__}: {self.name}\n'
           + (f' element: {self.element_name}' if self.element else '')
           + (f' system: {self.system}\n' if self.system else '')
           + (f' units: {self.units}\n' if self.units else '')
           + (f' distribution: {self.distribution}\n' if self.distribution else ''))
                
    
               